/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_async.c,v 1.11 2003/11/14 19:55:14 phargrov Exp $
 *
 *
 * This file deals with Phase1 threads/tasks.
 * The filename "cr_async.c" is legacy from when these were
 * known as "asynchronous threads/tasks".
 */


#include <linux/module.h>
#include <linux/sched.h>
#include <asm/uaccess.h>

#include "cr_module.h"

// cr_phase1_release(filp, priv)
//
// Remove Phase1 tasks from the task list when the file is closed.
// Note that the close might be implictly at exit and could also
// come from a task other than the one originally registered.
//
// XXX: could keep a hash mapping phase1_filp -> cr_task rather than
// walking the entire list each time.
void cr_phase1_release(struct file *filp, cr_pdata_t *priv)
{
	struct list_head *elem, *next;
	cr_task_t *cr_task;

	write_lock(&cr_task_lock);
	list_for_each_safe(elem, next, &cr_task_list) {
		cr_task = list_entry(elem, cr_task_t, task_list);
		if (cr_task->phase1_filp == filp) {
			cr_task->phase1_filp = NULL;
			cr_task->handler = NULL;
			__cr_task_put(cr_task);
		}
	}
	write_unlock(&cr_task_lock);
}

// cr_suspend(timeval)
//
// Wait for a checkpoint request to arrive.
// Argument is a struct timeval for a bounded wait, or NULL to block forever.
//
// Returns:
//	<0 on errors (including interrupted sleep)
//	0 otherwise
//
int cr_suspend(struct file *filp, void *arg)
{
	cr_task_t	*cr_task;
	struct timeval	timeval;
	unsigned long	timeout;
	int		retval = 0;

	CR_KTRACE_FUNC_ENTRY();

	cr_task = cr_task_get(current);
	if (!cr_task) {
		return -EINVAL;
	}

	// Fetch/decode the optional timeout from userspace
	timeout = MAX_SCHEDULE_TIMEOUT;
	if (arg) {
		retval = copy_from_user(&timeval, arg, sizeof(timeval));
		if (retval) {
			return retval;
		}

		timeout = timeval_to_jiffies(&timeval);
	} 

	// Block until signal is pending or cr_task->req is not NULL.
	for (;;) {
		set_current_state(TASK_INTERRUPTIBLE);

		if (cr_task->req || !timeout) {
			break;
		}

		if (signal_pending(current)) {
			break;
		}

		timeout = schedule_timeout(timeout);
	}
	current->state = TASK_RUNNING;

	if (arg) {
		int tmp;

		jiffies_to_timeval(timeout, &timeval);
		tmp = copy_to_user(arg, &timeval, sizeof(timeval));
		if (tmp) {
			return tmp;
		}
	}

	if (!cr_task->req && timeout) {
		// caught a signal other than the checkpoint trigger
		retval = -EINTR;
	}

	return retval;
}

// cr_phase1_register(file, arg)
//
// Registers the calling task as having a Phase1 checkpoint handler.
// Returns:
// 	0 on success
// 	-EAGAIN if a checkpoint is pending for this task
// 	-EEXIST if this task is already registered
// 	other < 0 on other error conditions
int cr_phase1_register(struct file *filp, void *arg)
{
	cr_task_t	*cr_task;
	int		retval;

	CR_KTRACE_FUNC_ENTRY();

	write_lock(&cr_task_lock);
	cr_task = __cr_task_get(current, 1);

	retval = -ENOMEM;
	if (!cr_task) {
		// Can't create the task list entry.
		goto out;
	}

	retval = -EAGAIN;
	if (cr_task->req) {
		// Can't register a Phase1 handler task
		// while a checkpoint is in progress.
		goto out;
	}

	retval = -EEXIST;
	if (cr_task->phase1_filp) {
		// Duplicate registration
		goto out;
	}

	retval = 0;
	cr_task->phase1_filp = filp;
	cr_task->handler = arg;

out:
	write_unlock(&cr_task_lock);

	return retval;
}
