/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_chkpt_req.c,v 1.84 2003/11/14 19:55:14 phargrov Exp $
 */

#include <linux/module.h>

#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/time.h>
#include <asm/uaccess.h>

#include "cr_module.h"

// Interval for the watchdog to wakeup HZ = 1s
#define CR_CHKPT_WD_INTERVAL	(HZ)

// release_request()
//
// Release a request structure, free if reference count drops to zero.
//
// Returns non-zero if the request has actually been freed.
//
// No locks needed
static int release_request(cr_chkpt_req_t *req)
{
	int result = atomic_dec_and_test(&req->ref_count);

	if (result) {
		cr_dest_free(&req->dest);
		kfree(req);
		MOD_DEC_USE_COUNT;
	}

	return result;
}

// __delete_from_req(req, cr_task)
//
// Remove a given task from the request list and try to free resources.
// Returns non-zero if the request list becomes empty.
//
// Must be called w/ cr_task_lock held for writing.
// Must be called w/ req->lock held for writing or when exclusive
// access to the request list is otherwise guaranteed.
static int __delete_from_req(cr_chkpt_req_t *req, cr_task_t *cr_task)
{
	int empty;

	list_del_init(&cr_task->req_list);
	empty = list_empty(&req->tasks);
	if (empty) {
		wake_up(&req->wait);
	}
	cr_task->req = NULL;
	__cr_task_put(cr_task);

	return empty;
}

// delete_from_req(req, cr_task)
//
// Remove a given task from the request list and try to free resources.
// Returns non-zero if the request list becomes empty.
//
// Call w/o holding locks
static int delete_from_req(cr_chkpt_req_t *req, cr_task_t *cr_task)
{
	int empty;

	write_lock(&cr_task_lock);
	write_lock(&req->lock);
	empty = __delete_from_req(req, cr_task);
	write_unlock(&req->lock);
	write_unlock(&cr_task_lock);

	return empty;
}

// delete_dead_task(req, cr_task)
//
// Remove a dead (or dying) task from the request
//
// Must be called w/ cr_task_lock held for writing.
// Must be called w/ req->lock held for writing or when exclusive
// access to the request list is otherwise guaranteed.
//
// XXX: Should we be raising some error flag in the req?
static void delete_dead_task(cr_chkpt_req_t *req, cr_task_t *cr_task)
{
	if (cr_task->phase1_filp) {
		cr_barrier_break(&req->phase_barrier);
	}
	cr_barrier_break(&req->predump_barrier);
	cr_barrier_break(&req->postdump_barrier);
	__delete_from_req(req, cr_task);
	(void)release_request(req);
}

// check_done(req)
// 
// Returns non-zero if the request can be reaped.
//
// Call w/o holding the locks
static int check_done(cr_chkpt_req_t *req)
{
	int retval;

	write_lock(&req->lock);
	retval = list_empty(&req->tasks);
	write_unlock(&req->lock);

	return retval;
}

// add_watchdog(req)
//
// Add the timer for the watchdog, taking care not to schedule past
// the request's expiration time.
// XXX: no expiration yet!
//
// Must be called w/ req->lock held for writing or when exclusive
// access to the request list is otherwise guaranteed.
static void add_watchdog(cr_chkpt_req_t *req)
{
	unsigned long expires;
	
	expires = jiffies + CR_CHKPT_WD_INTERVAL;
	if (req->has_expiration && time_after(expires, req->expiration)) {
		expires = req->expiration;
	}

	req->watchdog.timer.expires = expires;
	add_timer(&req->watchdog.timer);
}

// watchdog_timer(req)
//
// Timer which schedules a watchdog to run in process context
static void watchdog_timer(unsigned long arg)
{
	schedule_task((struct tq_struct *)arg);
}

// watchdog_task(req)
//
// Called periodically to cleanup after zombies and to enforce
// the time limit on a checkpoint request.
//
// XXX: no time limit yet
static void watchdog_task(void *arg)
{
	cr_chkpt_req_t *req = arg;
	struct list_head *elem, *next;
	cr_task_t *cr_task;
	int empty;

	write_lock(&cr_task_lock);
	write_lock(&req->lock);

	list_for_each_safe(elem, next, &req->tasks) {
		cr_task = list_entry(elem, cr_task_t, req_list);
		if (cr_task->task->state == TASK_ZOMBIE) {
			CR_WARN("%s: process %d exited with code %d "
				"and signal %d during checkpoint",
				__FUNCTION__,
				cr_task->task->pid,
				(cr_task->task->exit_code & 0xff00) >> 8,
				cr_task->task->exit_code & 0x7f);
			// Note that we hold one reference, thus we are
			// certain that this call won't destroy req.
			delete_dead_task(req, cr_task);
		}
	}
	empty = list_empty(&req->tasks);

	if (!empty) {
		if (req->has_expiration &&
		    time_after_eq(jiffies, req->expiration)) {
			CR_INFO("%s: timeout detected", __FUNCTION__);
			// XXX: handle timeout case here
#if 1 // XXX: temporary alternative
			req->has_expiration = 0;
			add_watchdog(req);
#endif
		} else {
			add_watchdog(req);
		}
	}

	write_unlock(&req->lock);
	write_unlock(&cr_task_lock);

	if (empty) {
		release_request(req);
	}
}

// alloc_request()
//
// Allocate and initialize a request structure
//
// No locks needed
static cr_chkpt_req_t * alloc_request(void)
{
	cr_chkpt_req_t *req;

	req = kmalloc(sizeof(*req), GFP_ATOMIC);
	if (req) {
		MOD_INC_USE_COUNT;
		memset(req, 0, sizeof(*req));
		atomic_set(&req->ref_count, 1);
		INIT_LIST_HEAD(&req->tasks);
		init_waitqueue_head(&req->wait);
		req->lock = RW_LOCK_UNLOCKED;
		req->result = 0;
		req->has_expiration = 0;
		cr_barrier_init(&req->phase_barrier, 0);
		cr_barrier_init(&req->predump_barrier, 0);
		cr_barrier_init(&req->postdump_barrier, 0);
		INIT_TQUEUE(&req->watchdog.task, &watchdog_task, req);
		init_timer(&req->watchdog.timer);
		req->watchdog.timer.function = &watchdog_timer;
		req->watchdog.timer.data = (unsigned long)&req->watchdog.task;
		init_MUTEX(&req->serial_mutex);
        	set_bit(0, &req->wrote_header);
        	set_bit(0, &req->wrote_linkage);
#if HAVE_RH9_SIGNAL_STRUCT
        	set_bit(0, &req->wrote_signal_struct);
#endif
        	set_bit(0, &req->wrote_fs_struct);
        	set_bit(0, &req->wrote_files_struct);
	}

	return req;
}

// bad_perms(task)
//
// Verify that the 'current' task is permitted to checkpoint the
// indicated task.  The checks are based on ensuring that the
// invoking user "owns" the task and will not expose sensitive data.
//
// Must call w/ tasklist_lock held to prevent task from disappearing.
//
// Returns 0 if OK.
// Returns non-zero error code if permission should be denied.
//
static int bad_perms(struct task_struct *task)
{
	int result = 0;

	// We are always supposed to check suser last.
	task_lock(task);
	if ((!(task->mm && task->mm->dumpable) ||
	     ((current->euid != task->suid) && (current->euid != task->uid) &&
	      (current->uid  != task->suid) && (current->uid  != task->uid)))
	    && !suser()) {
		result = -EPERM;
	}
	task_unlock(task);

	return result;
}

// add_task(req, task)
//
// This function first verfies permissions to checkpoint the task.
// This function next records the req->task and task->req associations.
// The req->task association in kept by the task_list field of the req.
// The task->req association is kept by searching the list of requests,
// but could be put in a hash later.
//
// req->lock is not needed here because we are sole user.
// cr_task_lock must be held for the lookups.
//
// Returns 0 in the normal case.
// Returns non-zero error code on failure, such as out-of-memory.
//
static int add_task(cr_chkpt_req_t *req, struct task_struct *task)
{
	cr_task_t *cr_task;
	int result;

	result = bad_perms(task);
	if (result) {
		// Bad permissions
		goto out;
	}

	cr_task = __cr_task_get(task, 1);

	result = -ENOMEM;
	if (!cr_task) {
		// No memory
		goto out;
	}

	result = -EBUSY;
	if (cr_task->req) {
		// Overlapping request
		goto out;
	}

	// XXX: should also check for overlapping w/ RESTART handlers.
	// Specifically the restart request handling code should make
	// some entry in the cr_task list and should remove that entry
	// in the HAND_RSTRT ioctl() code.

	result = 0;
	if (cr_task->phase1_filp) {
		req->phase_barrier.count += 1;
	}
	req->predump_barrier.count += 1;
	req->postdump_barrier.count += 1;
	cr_task->req = req;
	list_add_tail(&cr_task->req_list, &req->tasks);
	atomic_inc(&req->ref_count);

out:
	return result;
}

// build_req_task(req, target)
//
// Locates a single task by its pid verifies that the current task
// has permissions necessary to checkpoint them.
//
// Returns 0 if the task has been located and permissions verified.
// Returns non-zero error code otherwise.
//
static int build_req_task(cr_chkpt_req_t *req, pid_t target)
{
	struct task_struct *task;
	int result = 0;

	read_lock(&tasklist_lock);
	task = target ? find_task_by_pid(target)
		      : current; // Default target is self
	if (task) {
		result = add_task(req, task);
	}
	read_unlock(&tasklist_lock);

	req->checkpoint_scope = CR_SCOPE_TASK;

	return result;	
}

// build_req_proc(req, target)
//
// Locates the tasks which comprise a potentially multi-threaded
// process and verifies that the current task has permissions
// necessary to checkpoint them.
// A multi-threaded process is identified by the pid of any thread.
//
// Returns 0 if the tasks have been located and permissions verified.
// Returns non-zero error code otherwise.
//
static int build_req_proc(cr_chkpt_req_t *req, pid_t target)
{
	struct task_struct *task;
	struct task_struct *proc;
	int result = 0;
	CR_FOR_EACH_TASK_VARS;

	read_lock(&tasklist_lock);
	proc = target ? find_task_by_pid(target)
		      : current; // Default target is self
	/* Get all the tasks in the process */
	if (proc) {
#if HAVE_2_4_TASK_LINKAGE
		CR_FOR_EACH_TASK_PROC(proc, task) {
			result = add_task(req, task);
			if (result) {
				break;
			}
		}
#elif HAVE_RH9_TASK_LINKAGE
        	if (atomic_read(&proc->signal->count) == atomic_read(&proc->mm->mm_users)) {
			/* NEW (nptl) pthreads (or single threaded) */
			CR_FOR_EACH_TASK_TGID(proc->tgid, task) {
				result = add_task(req, task);
				if (result) {
					break;
				}
			}
		} else {
			/* OLD (linuthreads) pthreads */
			CR_FOR_EACH_TASK_PROC(proc, task) {
				result = add_task(req, task);
				if (result) {
					break;
				}
			}
		}
#else
    #error
#endif
	}
	read_unlock(&tasklist_lock);

	req->checkpoint_scope = CR_SCOPE_PROC;
	return result;
}

// build_req_pgrp(req, target)
//
// Locates the tasks with a given process group id and verifies that the
// current task has permissions necessary to checkpoint them.
//
// Returns 0 if the tasks have been located and permissions verified.
// Returns non-zero error code otherwise.
//
static int build_req_pgrp(cr_chkpt_req_t *req, pid_t target)
{
	struct task_struct *task;
	int result = 0;
	CR_FOR_EACH_TASK_VARS;

	if (!target) {
		// Default target is own pgrp
		target = current->pgrp;
	}

	read_lock(&tasklist_lock);
	CR_FOR_EACH_TASK_PGID(target, task) {
		result = add_task(req, task);

		if (result) {
			break;
		}
	}
	read_unlock(&tasklist_lock);

	req->checkpoint_scope = CR_SCOPE_PGRP;

	return result;
}

// build_req_sess(req, target)
//
// Locates the tasks with a given session id and verifies that the
// current task has permissions necessary to checkpoint them.
//
// Returns 0 if the tasks have been located and permissions verified.
// Returns non-zero error code otherwise.
//
static int build_req_sess(cr_chkpt_req_t *req, pid_t target)
{
	struct task_struct *task;
	int result = 0;
	CR_FOR_EACH_TASK_VARS;

	if (!target) {
		// Default target is own session
		target = current->session;
	}

	read_lock(&tasklist_lock);
	CR_FOR_EACH_TASK_SID(target, task) {
		result = add_task(req, task);

		if (result) {
			break;
		}
	}
	read_unlock(&tasklist_lock);

	req->checkpoint_scope = CR_SCOPE_SESS;

	return result;
}

// cr_chkpt_req(user_req)
//
// Processes a checkpoint request received from user space.
// Dispatches to build_req_*() to build a list of tasks.
// Calls cr_trigger_all() to trigger the checkpoint.
//
// Returns 0 on success, or a negative error code on failure.
int cr_chkpt_req(struct file *filp, void *arg)
{
	struct cr_chkpt_req	 ureq;
	cr_pdata_t		*priv;
	cr_chkpt_req_t		*req;
	int result = 0;

	CR_KTRACE_FUNC_ENTRY();

	// Check that we don't have a request on this fd already
	priv = filp->private_data;
	if (priv->chkpt_req) {
		return -EAGAIN;
	}

	// Allocate the kernel's structure to track the request
	req = alloc_request();
	if (!req) {
		return -ENOMEM;
	}

	// Copy the user's request
	if (copy_from_user(&ureq, arg, sizeof(ureq))) {
		kfree(req);
		return -EFAULT;
	}

	// XXX: Also copy the options if not NULL
	req->dump_format = ureq.dump_format;
	req->signal = ureq.signal;
	req->flags = ureq.flags;

	// Output for debugging
	CR_INFO("checkpoint request:");
	CR_INFO(" target:  %u", ureq.cr_target);
	CR_INFO("  scope:  %u", ureq.cr_scope);
	CR_INFO("   secs:  %u", ureq.cr_secs);
	CR_INFO("   opts:  %p", ureq.cr_opts);
	CR_INFO("     fd:  %d", ureq.cr_fd);

	// Validate the destination file descriptor
	result = cr_dest_init(&req->dest, ureq.cr_fd, filp);

	// Hold the lock needed to ensure the request is constructed
	// atomically w.r.t. registration of Phase[12] checkpoint tasks
	// and other checkpoint requests.
	write_lock(&cr_task_lock);

	// Build the list of tasks
	if (!result) {
		switch(ureq.cr_scope) {
		case CR_SCOPE_TASK:
			result = build_req_task(req, ureq.cr_target);
			break;

		case CR_SCOPE_PROC:
			result = build_req_proc(req, ureq.cr_target);
			break;

		case CR_SCOPE_PGRP:
			result = build_req_pgrp(req, ureq.cr_target);
			break;

		case CR_SCOPE_SESS:
			result = build_req_sess(req, ureq.cr_target);
			break;

		default:
			result = -EINVAL;
		}
	}

	if (!result) {
		// Validate the task list
		if (list_empty(&req->tasks)) {
			// Fail if list is empty
			result = -ESRCH;
		} else if (req->flags & CR_CHKPT_PROHIBIT_SELF) {
			struct list_head *elem;
			cr_task_t *cr_task;

			// If the request says to prohibit "self", then scan
			// the list for "current", returning EDEADLK if found.
			//
			// req->lock not needed because we are sole user of req
			list_for_each(elem, &req->tasks) {
				cr_task = list_entry(elem, cr_task_t, req_list);
				if (cr_task->task == current) {
					result = -EDEADLK;
					break;
				}
			}
		}
	}

	// Either cleanup or finalize
	if (result) {
		struct list_head *elem, *next;
		cr_task_t *cr_task;

		// XXX: is this enough cleanup?
		//
		// req->lock not needed because we are sole user of req
		list_for_each_safe(elem, next, &req->tasks) {
			cr_task = list_entry(elem, cr_task_t, req_list);
			__delete_from_req(req, cr_task);
			// Note that we hold one reference, thus we are
			// certain release_request() won't destroy req.
			(void)release_request(req);
		}

		write_unlock(&cr_task_lock);

		release_request(req);
	} else {
		// Send the triggers
		cr_trigger_phase1(req);

		// Release the lock
		write_unlock(&cr_task_lock);

		// add watchdog timer, which constitutes one reference:
		if (ureq.cr_secs && (ureq.cr_secs < (MAX_JIFFY_OFFSET / HZ))) {
			req->expiration = jiffies + HZ * ureq.cr_secs;
			req->has_expiration = 1;
		} else {
			// Either the requester specified no expiration or
			// the requested expiration is too large to handle.
			req->has_expiration = 0;
		}
		add_watchdog(req);
		// watchdog gets its own reference
		atomic_inc(&req->ref_count);

		priv->chkpt_req = req;
	}


	return result;
}


/**
 * cr_abort_checkpoint
 * @filp: The file on which the ioctl was received.
 * @flags:  abort type
 *
 * Returns:   0 if checkpoint successfully aborted.
 */
int cr_abort_checkpoint(struct file *filp, unsigned long flags)
{
	cr_chkpt_req_t *req;
	cr_task_t *cr_task;
	int result = 0;

	CR_KTRACE_FUNC_ENTRY("flags=0x%lx", flags);

	/* Lookup this task, and get request from it */
	cr_task = cr_task_get(current);
	if (!cr_task) {
		return -ESRCH;
	}
	req = cr_task->req;
	if(!req) {
		cr_task_put(cr_task);
		return -ESRCH;
	}
	write_lock(&req->lock);
	
	/* Set status flag of checkpoint to aborted: causes other tasks in
	 * checkpoint to abort by end of predump barrier. */
	if (flags & CR_CHECKPOINT_TEMP_FAILURE) {
		result = req->result = -CR_ETEMPFAIL;
	} else if (flags & CR_CHECKPOINT_PERM_FAILURE) {
		result = req->result = -CR_EPERMFAIL;
	}

	write_unlock(&req->lock);
	cr_task_put(cr_task);
	return result;
}



// cr_chkpt_task_complete(cr_task)
// 
// Routine to release a given cr_ckpt_req_t.
// Called when a task has completed checkpointing itself.
void cr_chkpt_task_complete(cr_task_t *cr_task)
{
	cr_chkpt_req_t *req = cr_task->req;

	delete_from_req(req, cr_task);
	release_request(req);
}

// cr_chkpt_req_release(filp, priv)
//
// Routine to release checkpoint request structures
// when a task closes the file (or exits).
//
// Must be called BEFORE cr_phase[12]_release
void cr_chkpt_req_release(struct file *filp, cr_pdata_t *priv)
{
	cr_chkpt_req_t *req;
	cr_task_t *cr_task;

	// If we are a "target" of a request with registered handlers,
	// then cleanup from what appears to be an unexpected exit/crash.
	// Note that we can't do this for filp's which don't have handlers
	// because cr_ctrl_fd() can validly close() an fd with no handlers.
	cr_task = cr_task_get(current);
	if (cr_task) {
		if (cr_task->req &&
		    (cr_task->phase1_filp || cr_task->phase2_filp)) {
			delete_dead_task(cr_task->req, cr_task);
		}
		cr_task_put(cr_task);
	}

	// If we are a requester, release the outstanding request.
	req = priv->chkpt_req;
	if (req) {
		(void)release_request(req);
	}
}

// cr_chkpt_done(timeval)
//
// Test/wait for a checkpoint request to complete.
// Argument is:
//    Pointer to a (struct timeval) for a bounded wait
//    	if zero then we take care not to schedule().
//    NULL for an unbounded wait
//
// Returns:
//	<0 on errors (including interrupted sleep)
//	>0 if complete
//	0 if not complete within the timeout (if any)
//
// XXX: Note that we can get stuck if a task rudely exits/dies while we
// are waiting.  We should consider waking up to reap such tasks.  That
// should be taken care of when a checkpoint time limit is added.
//
// XXX: consider ERESTARTSYS/ERESTARTNOHAND in place of EINTR?
int cr_chkpt_done(struct file *filp, void *arg)
{
	cr_pdata_t		*priv;
	cr_chkpt_req_t		*req;
	struct timeval		timeval;
	DECLARE_WAITQUEUE(wait, current);
		
	unsigned long timeout;
	int retval = 0;

	CR_KTRACE_FUNC_ENTRY();

	priv = filp->private_data;
	if (!priv) {
		return -EINVAL;
	}
	req = priv->chkpt_req;
	if (!req) {
		return -EINVAL;
	}

	// Fetch/decode the optional timeout from userspace
	timeout = MAX_SCHEDULE_TIMEOUT;
	if (arg) {
		retval = copy_from_user(&timeval, arg, sizeof(timeval));
		if (retval) {
			return retval;
		}

		timeout = timeval_to_jiffies(&timeval);
	} 

	add_wait_queue(&req->wait, &wait);
	for (;;) {
		set_current_state(TASK_INTERRUPTIBLE);

		retval = check_done(req);

		if (retval || !timeout || signal_pending(current)) {
			break;
		}

		timeout = schedule_timeout(timeout);
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&req->wait, &wait);

	// Copy remaining time back to userspace if applicable
	if (arg) {
		jiffies_to_timeval(timeout, &timeval);
		copy_to_user(arg, &timeval, sizeof(timeval));
		// like select() we fail silently if user timeval is R/O
	}

	// Check if we were interrupted
	if (!retval && timeout) {
		retval = -EINTR;
	}

	return retval;
}

// cr_chkpt_reap()
//
// Reap a completed checkpoint.
// Returns:
// 	req->result on success
// 	-EINVAL if no un-reaped checkpoint request is associated w/ this fd
// 	-EAGAIN if the request is not completed

//
// XXX: should return some status information stored in the request structure.
int cr_chkpt_reap(struct file *filp)
{
	cr_pdata_t		*priv;
	cr_chkpt_req_t		*req;
	int			retval;	

	CR_KTRACE_FUNC_ENTRY();

	retval = -EINVAL;
	priv = filp->private_data;
	if (priv) {
		req = priv->chkpt_req;
		if (req) {
			retval = -EAGAIN;
			if (check_done(req)) {
				priv->chkpt_req = NULL;
				retval = req->result;
				(void)release_request(req);
			}
		}
	}

	return retval;
}
