/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_dump_self.c,v 1.70 2003/11/14 19:55:14 phargrov Exp $
 */

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/file.h>
#include <linux/slab.h>

#include "vmadump.h"
#include "cr_module.h"

#include "cr_context.h"

// Create a core file
static int cr_do_coredump(cr_chkpt_req_t *req, struct file *filp, int shared)
{
	int result = -EBADF;

	if (shared) {
		down(&req->dest.mutex);
	} 

	if (filp) {
		lock_kernel();
		result = -ENOSYS;
		if (current->binfmt && current->binfmt->core_dump) {
			unsigned long core_mask ;

			// core_mask has PF_DUMPCORE bit iff current did before
			core_mask = current->flags | ~PF_DUMPCORE;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,17))
			// In kernel versions prior to 2.4.17 you should hold
			// the mmap_sem for reading when invoking core dumps.
			down_read(&current->mm->mmap_sem);
			result = !current->binfmt->core_dump(0,
						    get_pt_regs(current),
						    filp);
			up_read(&current->mm->mmap_sem);
#else
			// In kernel versions 2.4.17 and up the core dump
			// method handles the necessary locking.
			result = !current->binfmt->core_dump(0,
							get_pt_regs(current),
							filp);
#endif

			// restore PF_DUMPCORE bit to its previous state
			current->flags &= core_mask;
		}
		unlock_kernel();
	}
 
	if (shared) {
		up(&req->dest.mutex);
	} 

	return result;
}

/*
 * XXX:  Fix to return number of bytes written
 */
static int cr_save_linkage(cr_chkpt_req_t *req, struct file *filp)
{
    struct list_head *l;
    struct cr_context_tasklinkage *linkage;
    int linkage_size;
    cr_task_t *cr_task;
    struct task_struct *task;
    struct task_struct *inittask;
    int task_count;
    int task_num;
    int retval;

    /* first count to find out how many tasks there are */
    read_lock(&req->lock);
    task_count = 0;
    list_for_each(l, &req->tasks) {
        ++task_count;
    }
    read_unlock(&req->lock);

    /* check task count, just ensure non-zero and that we're not bigger 
     * than 128kB (for kmalloc) */
    linkage = NULL;
    linkage_size = sizeof(*linkage) * task_count;
    if ((task_count > 0) && linkage_size < 131072) {
        /* allocate a cr_context_linkage large enough to hold all of the
         * available tasks */
        linkage = kmalloc(linkage_size, GFP_KERNEL);
    }
    if (linkage == NULL) {
        retval = -ENOMEM;
	goto out;
    }

    /* XXX:  The barrier code should be modified so that we don't have
     * to glean this information from the cr_task task list...  */

    /* XXX:  There is a race here.  Theoretically, someone could add
     * or remove a task while we're not holding req->lock...  Probably
     * should punt, force maximum memory allocation early on, and then
     * just write out the number of tasks scanned. */

    read_lock(&tasklist_lock);
    inittask = find_task_by_pid(1);
    read_lock(&req->lock);
    /* loop over each task, and copy important fields into the linkage */
    task_num = 0;
    list_for_each(l, &req->tasks) {
        cr_task = list_entry(l, cr_task_t, req_list);
	if (cr_task == NULL) {
	    CR_KTRACE_BADPARM("NULL cr_task found on task list!");
	    retval = -EINVAL;
	    goto out_nulltask;
	}
	task = cr_task->task;

	/* XXX: can a task that has exited still be on the task list?  If so,
	 * how would we ever now? 
	 *   -- Must we even care?  Even if the task_struct is garbage, it
	 *   only matters we'll build a new process around it... 
	 */
	
	linkage[task_num].myptr       = task;
	linkage[task_num].real_parent = CR_REAL_PARENT(task);
	linkage[task_num].parent      = CR_PARENT(task);

	/* Handle init specially */
	if (linkage[task_num].real_parent == inittask) {
	    linkage[task_num].real_parent = CR_PARENT_INIT;
	}
	if (linkage[task_num].parent == inittask) {
	    linkage[task_num].parent = CR_PARENT_INIT;
	}

#if HAVE_2_4_TASK_LINKAGE
	/* Nonsense to restore this, but we'll use it on the restart side */
	linkage[task_num].thread_group = task->thread_group;
#elif HAVE_RH9_TASK_LINKAGE
        linkage[task_num].group_leader = task->group_leader;
        linkage[task_num].clear_child_tid = task->clear_child_tid;
#else
	#error
#endif

	linkage[task_num].pid     = task->pid;
	linkage[task_num].pgrp    = task->pgrp;
	linkage[task_num].tty_old_pgrp = task->tty_old_pgrp;
	linkage[task_num].session = task->session;
	linkage[task_num].tgid    = task->tgid;
	linkage[task_num].leader  = task->leader;
	linkage[task_num].exit_signal  = task->exit_signal;
	++task_num;
    }

out_nulltask:
    read_unlock(&req->lock);
    read_unlock(&tasklist_lock);

    /* This will catch the goto, and a possible race condition */
    if (task_num < task_count - 1) {
        CR_KTRACE_BADPARM("Warning: wrong number of tasks counted: task_count = %d, task_num = %d", 
	    task_count, task_num);
	retval = -EINVAL;
    } else {
        /* write the length of the linkage structure to the context file */
        retval = cr_kwrite(filp, &linkage_size, sizeof(linkage_size));
	if (retval < sizeof(linkage_size)) {
            CR_KTRACE_LOW_LVL("cr_kwrite wrote %d of %d bytes", retval, 
			      sizeof(linkage_size));
            /* XXX: This is an error...  we should abort! */
	} else {
	    /* now write the linkage structure */
            retval = cr_kwrite(filp, linkage, linkage_size);
	    if (retval < linkage_size) {
                CR_KTRACE_LOW_LVL("cr_kwrite wrote %d of %d bytes", 
				  retval, linkage_size);
	    }
	}
    }

    kfree(linkage);
out:
    return retval;
}

#if CR_FILES_WORK
static struct file *
cr_dest_gett(cr_chkpt_req_t *req, struct task_struct *task, cr_obj_t what)
{
    int shared = -1;

    return cr_dest_get(&req->dest, &shared);
    CR_KTRACE_LOW_LVL("shared = %d", shared);
}

static void
cr_dest_putt(cr_chkpt_req_t *req, struct file *filp)
{
    cr_dest_put(&req->dest, filp);
}

/*
 * cr_fputs
 * 
 * Slightly different than fputs, since it writes the NULL also, and refuses to
 * write anything larger than a PAGE_SIZE
 * 
 * Returns number of bytes written on success.
 */
static int
cr_fputs(char *buf, struct file *filp)
{
    int ret;
    int wrote=0;
    int len;

    ret = -EIO;

    /* we actually want to know when someone wrote NULL to an area, so we
     * can restore the NULL on restart. */
    if (!buf) {
        /* magic length of 0 for NULL buf */
        len = 0;
    } else {
        /* we write out the length WITH the null termination. */
        len = strlen(buf)+1;
    }

    /* compare l+1 to avoid type issue wrt promotion of len+1 to unsigned
     * long in comparison against PAGE_SIZE */
    if (len < 0 || len+1 > PAGE_SIZE+1) {
        CR_ERR("cr_fputs:  String length (%d) out of bounds.", len);
        ret = -EINVAL;
        goto out;
    }

    ret = cr_kwrite(filp, &len, sizeof(len));
    wrote += ret;
    if (ret != sizeof(len)) {
        CR_ERR("cr_fputs: Bad file write");
        ret = -1;
        goto out;
    }
  
    if (len > 0) {
        ret = cr_kwrite(filp, buf, len);
        wrote += ret;
        if (ret != len) {
            CR_ERR("cr_fputs: Bad file write");
            ret = -1;
            goto out;
        }
    }

    ret = wrote;

out:
    return ret;
}

static inline char *
cr_getpath(struct dentry *dentry, struct vfsmount *mnt, char *buf, int size)
{
    struct dentry *d;
    struct vfsmount *m;
    char *path;

    path = NULL;

    if (dentry==NULL) {
        CR_WARN("dentry is NULL!");
	goto out;
    }
    if (mnt==NULL) {
        CR_WARN("mnt is NULL!");
	goto out;
    }

    d = dget(dentry);
    m = mntget(mnt);

    if (d==NULL) {
        CR_WARN("d dentry is NULL!");
	goto out;
    }
    if (m==NULL) {
        CR_WARN("m vfsmount is NULL!");
	goto out;
    }

    path = d_path(d, m, buf, size);
    dput(d);
    mntput(m); 

out:
    return path;
}

/*
 * XXX: change from PAGE_SIZE to PATH_MAX?
 */
static int 
cr_save_fs_struct(cr_chkpt_req_t *req)
{
    struct dentry *dentry;
    struct vfsmount *mnt;
    int retval;
    char *path;
    int umask;
    void *buf;
    struct file *filp;

    buf = (void *) __get_free_page(GFP_KERNEL);
    if (!buf) {
        retval = -ENOMEM;
        goto out_nomem;
    }

    filp = cr_dest_gett(req, current, cr_fs_obj);
    if (!filp) {
        retval = -EBADF;
        goto out_nofile;
    }

    read_lock(&current->fs->lock);
    umask = current->fs->umask;
    read_unlock(&current->fs->lock);

    /* save umask */
    retval = cr_kwrite(filp, &umask, sizeof(umask));
    if (retval < sizeof(umask)) {
        CR_ERR("umask: Bad file write!");
        goto out_writerr;
    }

    /* save root */
    read_lock(&current->fs->lock);
    dentry = current->fs->root;
    mnt = current->fs->rootmnt;
    read_unlock(&current->fs->lock);
    path = cr_getpath(dentry, mnt, buf, PAGE_SIZE);
    retval = cr_fputs(path, filp);
    if (retval < 0) {
        CR_ERR("root: Bad file write!");
        goto out_writerr;
    }

    /* save altroot -- can be NULL */
    read_lock(&current->fs->lock);
    dentry = current->fs->altroot;
    mnt = current->fs->altrootmnt;
    read_unlock(&current->fs->lock);
    if (dentry) {
        path = cr_getpath(dentry, mnt, buf, PAGE_SIZE);
    } else {
        path = NULL;
    }
    retval = cr_fputs(path, filp);
    if (retval < 0) {
        CR_ERR("altroot: Bad file write!");
        goto out_writerr;
    }

    /* save pwd */
    read_lock(&current->fs->lock);
    dentry = current->fs->pwd;
    mnt = current->fs->pwdmnt;
    read_unlock(&current->fs->lock);
    path = cr_getpath(dentry, mnt, buf, PAGE_SIZE);
    retval = cr_fputs(path, filp);
    if (retval < 0) {
        CR_ERR("pwd: Bad file write!");
        goto out_writerr;
    }

out_writerr:
    cr_dest_putt(req, filp);
out_nofile:
    free_page((unsigned long) buf);
out_nomem:
    return retval;
}

/*
 * Saves the files_struct
 *
 * This looks silly, but we actually restore the values of max_fds array
 * later.  Since we know how many files are open, we don't have to guess
 * the way get_unused_fd does.
 */
static int cr_save_files_struct(cr_chkpt_req_t *req, struct files_struct *files)
{
    struct cr_files_struct cr_fs;
    struct file *c_filp;
    int retval=0;

    c_filp = cr_dest_gett(req, current, cr_files_obj);
    if (!c_filp) {
        CR_ERR("Unable to find destination file for cr_files_struct.");
        retval = -EINVAL;
        goto out_no_cfilp;
    }

    read_lock(files->file_lock);
    cr_fs.cr_obj = cr_files_obj; 
    cr_fs.cr_max_fds = files->max_fds;
    cr_fs.cr_next_fd = files->next_fd;
    cr_fs.cr_max_fdset = files->max_fdset;
    read_unlock(files->file_lock);

    retval = cr_kwrite(c_filp, &cr_fs, sizeof(cr_fs));
    if (retval != sizeof(cr_fs)) {
        CR_ERR("cr_fs: Bad file write! (err=%d)", retval);
        retval = -EIO;
        goto out_writerr;
    }

out_writerr:
    cr_dest_putt(req, c_filp);
out_no_cfilp:
    return retval;
}

/*
 * cr_get_fd_info
 *
 * Saves flags associated with the file descriptor.  (stored in files_struct)
 *
 * Must be called with the file_lock held!
 * 
 * Assumes that the file is open.  (descriptor refers to an open file in fd_array)
 */
static int
cr_get_fd_info(struct files_struct *files, int fd, struct cr_file_info *file_info)
{
    int i;

    if (FD_ISSET(fd, files->open_fds)) {
        file_info->fd      = fd;
        file_info->cloexec = FD_ISSET(fd, files->close_on_exec);
        file_info->is_open = FD_ISSET(fd, files->open_fds);
    } else {
        CR_WARN("cr_get_fd_info: Called on closed file!");
    }

    /* check for dups -- easiest here. */
    file_info->cr_dup = -1;
    for (i=0; i<fd; ++i) {
        if (files->fd_array[i] == files->fd_array[fd]) {
            file_info->cr_dup       = i;
            goto out;
        }
    }

out:
    return 0;
}

static int
cr_get_file_info(struct file *filp, struct cr_file_info *file_info)
{
    unsigned int mode;
    int retval=0;

    mode = filp->f_dentry->d_inode->i_mode;

    if (file_info->cr_dup != -1) {
        file_info->cr_file_type = cr_open_dup;
        goto out;
    }

    if (S_ISREG(mode)) {
        file_info->cr_file_type = cr_open_file;
    } else if (S_ISDIR(mode)) {
        file_info->cr_file_type = cr_open_directory;
    } else if (S_ISLNK(mode)) {
        file_info->cr_file_type = cr_open_link;
    } else if (S_ISFIFO(mode)) {
        file_info->cr_file_type = cr_open_fifo;
    } else if (S_ISSOCK(mode)) {
        file_info->cr_file_type = cr_open_socket;
    } else if (S_ISCHR(mode)) {
        file_info->cr_file_type = cr_open_chr;
    } else if (S_ISBLK(mode)) {
        file_info->cr_file_type = cr_open_blk;
    } else {
        /* completely unknown */
        file_info->cr_file_type = cr_unknown_file;
    }

out:
    return retval;
}

static int
cr_save_file_info(cr_chkpt_req_t *req, struct cr_file_info *file_info)
{
    struct file *filp;
    int retval;

    filp = cr_dest_gett(req, current, cr_file_obj);
    if (!filp) {
        retval = -EINVAL;
        goto out;
    }
    
    retval = cr_kwrite(filp, file_info, sizeof(*file_info));
    if (retval != sizeof(*file_info)) {
        retval = -EINVAL;
        CR_ERR("cr_save_file_info: Bad write!");
        goto out_badwrite;
    }

    retval = 0;

out_badwrite:
    cr_dest_putt(req, filp);
out:
    return retval;
}

static int
cr_save_open_file(cr_chkpt_req_t *req, struct file *filp)
{
    char *name;
    struct cr_open_file open_file;
    char *buf;
    int len;
    struct file *c_filp;                /* the context file */
    int retval;

    c_filp = cr_dest_gett(req, current, cr_file_obj);
    if (!c_filp) {
        retval = -EBADF;
        goto out;
    }

    buf = (char *) __get_free_page(GFP_KERNEL);
    if (!buf) {
        retval = -ENOMEM;
        goto out;
    }

    open_file.cr_type = cr_open_file_obj;
    /* XXX:  Fix to use better representation. */
    open_file.i_mode  = filp->f_dentry->d_inode->i_mode;
    open_file.f_flags = filp->f_flags;
    open_file.f_pos   = filp->f_pos;
    
    /* find the file name and length */
    name = cr_getpath(filp->f_dentry, filp->f_vfsmnt, buf, PAGE_SIZE);
    if (name == NULL) {
        CR_ERR("Bad file name!");
        retval = -EBADF;
        goto out_bad;
    }
    len = strlen(name);
    open_file.namelen = len;

    /* write out the open_file struct */
    retval = cr_kwrite(c_filp, &open_file, sizeof(open_file));
    if (retval != sizeof(open_file)) {
        CR_ERR("cr_save_open_file - Bad file write!");
        goto out_bad;
    }

    /* now write out the name */
    retval = cr_fputs(name, c_filp);
    if (retval < len) {
        CR_ERR("cr_save_open_file - Bad file write! (fputs returned %d)", retval);
        goto out_bad;
    }

    cr_dest_putt(req, c_filp);
    
out_bad:
    free_page((unsigned long) buf);
out:
    return retval;
}

static int 
cr_save_open_dir(cr_chkpt_req_t *req, struct file *filp)
{
    int retval = -ENOSYS;

    return retval;
}

static int 
cr_save_open_link(cr_chkpt_req_t *req, struct file *filp)
{
    int retval = -ENOSYS;

    return retval;
}

static int 
cr_save_open_fifo(cr_chkpt_req_t *req, struct file *filp)
{
    struct cr_dup cf_dup;
    struct file *cf_filp;
    int retval;

    retval = -EINVAL;
    if (!filp) 
      goto out;

    cf_filp = cr_dest_gett(req, current, cr_dup_obj);

    /* placeholder... just in case we need to do something later */
    cf_dup.cr_dupfd = -2;

    retval = cr_kwrite(cf_filp, &cf_dup, sizeof(cf_dup));
    if (retval != sizeof(cf_dup)) {
	retval = -EIO;
	CR_WARN("cr_kwrite error! = %d", retval);
	goto out_putt;
    }
    retval = 0;

out_putt:
    cr_dest_putt(req, cf_filp);
out:
    return retval;
}

static int 
cr_save_open_socket(cr_chkpt_req_t *req, struct file *filp)
{
    int retval = -ENOSYS;

    return retval;
}

/*
 * Right now, saves major and minor numbers.
 */
static int 
cr_save_open_chr(cr_chkpt_req_t *req, struct file *filp)
{
    struct cr_chrdev cf_chrdev;
    struct file *cf_filp;
    struct inode *inode;
    int retval;

    retval = -EINVAL;
    if (!filp) 
      goto out;

    cf_filp = cr_dest_gett(req, current, cr_chrdev_obj);

    inode = filp->f_dentry->d_inode;

    cf_chrdev.cr_type  = cr_open_chr;
    cf_chrdev.cr_major = MAJOR(inode->i_rdev); 
    cf_chrdev.cr_minor = MINOR(inode->i_rdev);

    retval = cr_kwrite(cf_filp, &cf_chrdev, sizeof(cf_chrdev));
    if (retval != sizeof(cf_chrdev)) {
	retval = -EIO;
	CR_WARN("cr_kwrite error! = %d", retval);
	goto out_write;
    }
    retval = 0;

out_write:
    cr_dest_putt(req, cf_filp);
out:
    return retval;
}

static int 
cr_save_open_blk(cr_chkpt_req_t *req, struct file *filp)
{
    int retval = -ENOSYS;

    return retval;
}

/*
 * dup info is associated with a file descriptor, not a file,
 * so it's saved in cr_save_file_info, and nothing is saved here,
 * other than a placeholder data structure.
 */
static int 
cr_save_open_dup(cr_chkpt_req_t *req, struct file *filp)
{
    struct cr_dup cf_dup;
    struct file *cf_filp;
    int retval;

    retval = -EINVAL;
    if (!filp) 
      goto out;

    cf_filp = cr_dest_gett(req, current, cr_dup_obj);

    /* placeholder... just in case we need to do something later */
    cf_dup.cr_type  = cr_open_dup;
    cf_dup.cr_dupfd = -1;

    retval = cr_kwrite(cf_filp, &cf_dup, sizeof(cf_dup));
    if (retval != sizeof(cf_dup)) {
	retval = -EIO;
	CR_WARN("cr_kwrite error! = %d", retval);
	goto out_write;
    }
    retval = 0;

out_write:
    cr_dest_putt(req, cf_filp);
out:
    return retval;
}

static int
cr_save_file_locks(cr_chkpt_req_t *req, struct file *filp)
{
    int retval = -ENOSYS;

    return retval;
}

static int
cr_save_all_files(cr_chkpt_req_t *req)
{
    int retval;
    int fd;
    struct cr_file_info file_info;
    struct file *filp;
    int max_fds;

    CR_KTRACE_FUNC_ENTRY("");

    read_lock(&current->files->file_lock);
    max_fds = current->files->max_fds;    
    read_unlock(&current->files->file_lock);

    /* save the files info */
    CR_KTRACE_HIGH_LVL("    ...files_struct");
    retval = cr_save_files_struct(req, current->files);
    if (retval < 0) {
        CR_ERR("cr_save_all_files: Error saving files_struct");
        goto out_nolocks;
    }

    CR_KTRACE_HIGH_LVL("    ...files");
    /* now save the per file info */
    read_lock(&current->files->file_lock);
    for (fd = 0; fd < current->files->max_fds && fd < max_fds; ++fd) {

        /* sanity check */
        if (fd > current->files->max_fdset) {
            /* XXX: we can probably handle this.  is it a good idea? */
            CR_WARN("fd larger than fdset!");
            read_unlock(&current->files->file_lock);
            retval = -EBADF;
            goto out_nolocks;
        }

        if (max_fds != current->files->max_fds) {
            CR_WARN("current->files->max_fds changed unexpectedly!");
        }

        /*
         * We have to do our own fget here to avoid a possible race on
         * file close.  (Probably impossible, but just to be on the safe
         * side.
         */

        /* loop around again if the file is not open */
        filp = fcheck(fd);
        if (! filp) {
            continue;
        }

        get_file(filp);

        CR_KTRACE_LOW_LVL("fd %d is open", fd);

	memset(&file_info, 0, sizeof(file_info));
	file_info.cr_type = cr_file_info_obj;

        /* now save data for this file descriptor */
        cr_get_fd_info(current->files, fd, &file_info);

        read_unlock(&current->files->file_lock);

        if (!file_info.is_open) {
            CR_WARN("Unopened file found in table???.");
            goto continue_fd;
        }

        /* write out the file header */
        retval = cr_get_file_info(filp, &file_info);
        if (retval) {
            retval = -EBADF;
            CR_ERR("Unable to determine file info!"); 
        }
        retval = cr_save_file_info(req, &file_info);
        if (retval < 0) {
            retval = -EIO;
            CR_ERR("Write error!");
            fput(filp);
            goto out_nolocks;
        }

        switch(file_info.cr_file_type) {
        case cr_open_file:
            CR_KTRACE_LOW_LVL("    ...%d is regular file.", fd);
            retval = cr_save_open_file(req, filp);
            break;
        case cr_open_directory:
            CR_KTRACE_LOW_LVL("    ...%d is open directory.", fd);
            retval = cr_save_open_dir(req, filp);
            break; 
        case cr_open_link:
            CR_KTRACE_LOW_LVL("    ...%d is open symlink.", fd);
            retval = cr_save_open_link(req, filp);
            break;
        case cr_open_fifo:
            CR_KTRACE_LOW_LVL("    ...%d is open fifo.", fd);
            retval = cr_save_open_fifo(req, filp);
            break;
        case cr_open_socket:
            CR_KTRACE_LOW_LVL("    ...%d is open socket.", fd);
            retval = cr_save_open_socket(req, filp);
            break;
        case cr_open_chr:
            CR_KTRACE_LOW_LVL("    ...%d is open character device.", fd);
            retval = cr_save_open_chr(req, filp);
            break;
        case cr_open_blk:
            CR_KTRACE_LOW_LVL("    ...%d is an open block device.", fd);
            retval = cr_save_open_blk(req, filp);
            break;
        case cr_open_dup:
            CR_KTRACE_LOW_LVL("    ...%d is dup of %d", fd, file_info.cr_dup);
            retval = cr_save_open_dup(req, filp);
            break;
        case cr_bad_file_type:
            /* fall through */
        default:
            retval = -EBADF;
            break;
        }

        if (retval < 0) {
            fput(filp);
            retval = -EBADF;
            CR_ERR("Unable to save open file!");
            goto out_nolocks;
        }

        /* Now write out any locks we had on the file. */
        CR_KTRACE_LOW_LVL("    ...locks on %d", fd);
        retval = cr_save_file_locks(req, filp);

        /* We did the fget() manually with the lock held. */
        fput(filp);

continue_fd:
        read_lock(&current->files->file_lock);
    }
    read_unlock(&current->files->file_lock);

    /* now write one last record to indicate that there are no files left. */
    memset(&file_info, 0, sizeof(file_info));
    file_info.cr_type = cr_file_info_obj;
    file_info.cr_file_type = cr_end_of_files;
    file_info.fd = -1;
    retval = cr_save_file_info(req, &file_info);
    if (retval < 0) {
        retval = -EIO;
        CR_ERR("Write error!");
        goto out_nolocks;
    }

out_nolocks:
    return retval;
}
#endif /* CR_FILES_WORK */

static int
cr_save_header(cr_chkpt_req_t *req, struct file *filp, int thread_count)
{
    int result;
    struct cr_context_header cf_head;

    result = 0;

    CR_KTRACE_LOW_LVL("Dumping header for %d threads", thread_count);
    cf_head.magic[0] = 'C';
    cf_head.magic[1] = 'R';
    cf_head.version = CR_CONTEXT_VERSION;
    cf_head.checkpoint_type = req->checkpoint_scope;
    cf_head.num_threads = thread_count;

    result = cr_kwrite(filp, &cf_head, sizeof(cf_head));

    if (result != sizeof(cf_head)) {
	result = -EIO;
	CR_WARN("cr_kwrite error! = %d", result);
	goto out;
    }
    result = 0;

  out:
    return result;
}

static int
needs_header(cr_chkpt_req_t *req) {
    return (test_and_clear_bit(0, &req->wrote_header));
}

static int
needs_linkage(cr_chkpt_req_t *req) {
    return (test_and_clear_bit(0, &req->wrote_linkage));
}

#if HAVE_RH9_SIGNAL_STRUCT
static int
needs_signal_struct(cr_chkpt_req_t *req) {
    return (test_and_clear_bit(0, &req->wrote_signal_struct));
}
#endif

#if CR_FILES_WORK
static int
needs_files_struct(cr_chkpt_req_t *req) {
    return (test_and_clear_bit(0, &req->wrote_files_struct));
}

static int
needs_fs_struct(cr_chkpt_req_t *req) {
    return (test_and_clear_bit(0, &req->wrote_fs_struct));
}
#endif

static int
needs_vmadump(cr_chkpt_req_t *req) {
    /* vmadump coordinates internally, so no need to worry. */
    return 1;
}

// Create a vmadump file
static int cr_do_vmadump(cr_chkpt_req_t *req, struct file *filp, int shared)
{
    struct vmadump_threaddump_args vmad_args;
    int thread_count = req->predump_barrier.entered;
    int result=0;

    vmad_args.num_threads = thread_count;
    vmad_args.fd = -1;
    vmad_args.flags = 0;

    /* Write out the header */
    down(&req->serial_mutex);
    if (needs_header(req)) {
        CR_KTRACE_HIGH_LVL("Preparing to dump %d threads", thread_count);
        result = cr_save_header(req, filp, thread_count);
        if (result < 0) {
            goto out_mutex;
        }
    }
    up(&req->serial_mutex);

    /* Now dump out the task linkage */
    down(&req->serial_mutex);
    if (needs_linkage(req)) {
        CR_KTRACE_LOW_LVL("Writing the linkage.");
        result = cr_save_linkage(req, filp);
        if (result < 0) {
	    goto out_mutex;
        }
    }
    up(&req->serial_mutex);

#if HAVE_RH9_SIGNAL_STRUCT
    /* XXX Need to save the actual shared signal struct(s).  For now
     * we just indicate if we have one or many for the threads. */
    down(&req->serial_mutex);
    if (needs_signal_struct(req)) {
	int is_shared = (atomic_read(&current->signal->count) > 1);
        CR_KTRACE_LOW_LVL("Writing shared signal struct(s)");

	result = cr_kwrite(filp, &is_shared, sizeof(is_shared));
        if (result < sizeof(is_shared)) {
	    result = -EIO;
            goto out_mutex;
        }
    }
    up(&req->serial_mutex);
#endif

    /* Now use vmadump to write out the state for the thread */
    if (needs_vmadump(req)) {
        CR_KTRACE_LOW_LVL("Writing vmadump.");
        result = vmadump_freeze_threads(filp, get_pt_regs(current), &vmad_args);
        if (result > 0) {
	    /* vmadump returns number of bytes written when successful */
	    result = 0;
        } else {
	    goto out;
        }
    }
    CR_KTRACE_LOW_LVL("finished vmadump");

#if CR_FILES_WORK
    /* dump fs_struct (cwd, umask, etc.) */
    down(&req->serial_mutex);
    if (needs_fs_struct(req)) {
        CR_KTRACE_HIGH_LVL("Writing the fs struct...");
        result = cr_save_fs_struct(req);
        if (result < 0) {
	    goto out_mutex;
        }
    }
    up(&req->serial_mutex);

    down(&req->serial_mutex);
    if (needs_files_struct(req)) {
        /* dump the open files */
        CR_KTRACE_HIGH_LVL("Writing the open file section...");
        result = cr_save_all_files(req);
        if (result < 0) {
	    goto out_mutex;
        }
    }
    up(&req->serial_mutex);
#endif

  out:
    result = 0; // XXX
    return result;

out_mutex:
    up(&req->serial_mutex);
    result = 0; // XXX
    return result;
}

// Dump out to the given file
//
// XXX: When we define an options struct, it will need to be passed in too

/*
 * XXX:  Might want to change this a little (lot) when we support multiple
 * context files.  Could be a separate format for directory set, etc.  i.e.
 * cr_format_contextfile, cr_format_contextdir?
 */
static int cr_do_dump(cr_chkpt_req_t *req)
{
	struct file *filp;
	int result = -EINVAL;
	int shared;

	filp = cr_dest_get(&req->dest, &shared);
	if (IS_ERR(filp)) {
		return PTR_ERR(filp);
	}

	switch(req->dump_format) {
	case cr_format_vmadump:
		result = cr_do_vmadump(req, filp, shared);
		break;

	case cr_format_coredump:
		result = cr_do_coredump(req, filp, shared);
		break;

	default:
		CR_KTRACE_BADPARM("unrecognized dump_format '%d'", 
				  req->dump_format);
		break;
	}

	cr_dest_put(&req->dest, filp);

	return result;
}


/*
 * Public functions
 */

/**
 * cr_dump_self - dump 'current'
 * @filp: The file on which the request was received.
 * @flag: Flags to modify behavior.
 *
 * DESCRIPTION:
 * Dumps the current process to a file descriptor supplied by the
 * requester.
 *
 * This is the function that the handlers will call into when
 * the CHECKPOINT portion of callbacks are finished.
 *
 * Currently the 'flags' argument is unused.
 */
int cr_dump_self(struct file *filp, unsigned long flags)
{
	cr_chkpt_req_t *req;
	cr_task_t *cr_task;
	int result = 0;

	CR_KTRACE_FUNC_ENTRY("flags=0x%lx", flags);

	// Lookup this task
	cr_task = cr_task_get(current);
	if (!cr_task) {
		// No matching task found.
		return -ESRCH;
	}

	// Find the matching request
	req = cr_task->req;
	if(!req) {
		// No matching request found.
		cr_task_put(cr_task);
		return -ESRCH;
	}

	// XXX: Once we define the options structure, this is the place
	// to merge the per-process options (attached to 'filp'?) and the
	// options passed by the requester.

	// If we have a Phase1 then start Phase2 when complete.
	if (cr_task->phase1_filp) {
		cr_barrier_enter(&req->phase_barrier);
		// All Phase1 threads are done, so trigger the Phase2 threads.
		cr_trigger_phase2(req);
	}

	/* Check to see if error/abort has occurred */
	read_lock(&req->lock);
	if ( (result = req->result)) {
		read_unlock(&req->lock);
		cr_barrier_break(&req->predump_barrier);
		goto cleanup;
	}
	read_unlock(&req->lock);

	// Synchronize to ensure all tasks have stopped running.
	// XXX: could do this at thread-group granularity in the future
	result = cr_barrier_enter(&req->predump_barrier);

	// The handlers have all responded, so we should not "expire"
	// So "first" task out of the barrier zeros req->has_expiration.
	if (result) {
		write_lock(&req->lock);
		req->has_expiration = 0;
		write_unlock(&req->lock);
	}

	/* check again for abort/error */
	read_lock(&req->lock);
	if ( (result = req->result)) {
		read_unlock(&req->lock);
		goto cleanup;
	}
	read_unlock(&req->lock);

	// Send the dump to the proper destination filp.
	result = cr_do_dump(req);

	// If we experienced an error for the first time then
	// save it in the request so the requester will learn of it.
	if (result && !req->result) {
		req->result = result;
	}

	// Synchronize to ensure all tasks have saved before we resume running.
	// XXX: could do this at thread-group granularity in the future
	cr_barrier_enter(&req->postdump_barrier);

	// Here is where we raise(SIG_STOP) etc
	send_sig_info(req->signal, NULL, current);

cleanup:
	if (!cr_task->handler) {
		// task has no handlers, so finish up now.
		cr_chkpt_task_complete(cr_task);
	} else {
		// cr_chkpt_task_complete() will run from the HAND_CONT ioctl.
	}

	// Release the task.  (XXX: THIS IS A BUG?  Do only if no handlers...)
	cr_task_put(cr_task);

	return result;
}

/**
 * cr_cont_complete
 * @filp: The file on which the ioctl was received.
 *
 * DESCRIPTION:
 * Implements the HAND_CONT ioctl() which indicates to the kernel that
 * the calling task has completed running its CONTINUE handlers.  Thus
 * we can prevent checkpoint requests from overlapping these handlers.
 *
 * This is the function that the handlers will call into when
 * the CONTINUE portion of callbacks are finished.
 *
 */
int cr_cont_complete(struct file *filp)
{
	cr_task_t *cr_task;

	CR_KTRACE_FUNC_ENTRY();

	// Lookup this task
	cr_task = cr_task_get(current);
	if (!cr_task) {
		// No matching task found.
		return -ESRCH;
	}

	// Find the matching request
	if(!cr_task->req) {
		// No matching request found.
		cr_task_put(cr_task);
		return -ESRCH;
	}

	// Release the request.
	cr_chkpt_task_complete(cr_task);

	// Release the task.
	cr_task_put(cr_task);

	return 0;
}
