/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_fops.c,v 1.31 2003/11/14 19:55:14 phargrov Exp $
 */

#include <linux/module.h>

#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/file.h>

#include "cr_module.h"

/**
 * ctrl_ioctl - ioctl() method for the C/R control node
 * @inode: (struct inode *) of the control node
 * @file: (struct file *) of the control node
 * @op: The operation/command to perform
 * @arg: The (optional) argument to the command as an unsigned long
 *
 * DESCRIPTION:
 * This function is the ioctl() method in the file_operations for
 * the C/R control node.  It acts as a dispatcher to the functions
 * which do the real work.
 */
static int ctrl_ioctl(struct inode *inode, struct file *file,
		      unsigned int op, unsigned long arg)
{
	CR_KTRACE_FUNC_ENTRY("op=%08x arg=0x%lx", op, arg);

	switch(op) {
	//
	// Calls from libcr:
	//
	case CR_OP_HAND_CHKPT:
		return cr_dump_self(file, arg);

	case CR_OP_HAND_ABORT:
		return cr_abort_checkpoint(file, arg);

	case CR_OP_HAND_CONT:
		return cr_cont_complete(file);

	case CR_OP_HAND_RSTRT:
		return cr_rstrt_barrier(file, arg);

	case CR_OP_HAND_SUSP:
		return cr_suspend(file, (void *)arg);

	case CR_OP_HAND_PHASE1:
		return cr_phase1_register(file, (void *)arg);

	case CR_OP_HAND_PHASE2:
		return cr_phase2_register(file, (void *)arg);

	//
	// Calls from cr_checkpoint:
	//
	case CR_OP_CHKPT_REQ:
		return cr_chkpt_req(file, (void *)arg);

	case CR_OP_CHKPT_DONE:
		return cr_chkpt_done(file, (void *)arg);

	case CR_OP_CHKPT_REAP:
		return cr_chkpt_reap(file);

	//
	// Calls from cr_restart:
	//
	case CR_OP_RSTRT_REQ:
		return cr_rstrt_request_restart(file, (void *)arg);

	case CR_OP_RSTRT_DONE:
		return cr_rstrt_done(file, (void *)arg);

	case CR_OP_RSTRT_REAP:
		return cr_rstrt_reap(file);

	case CR_OP_RSTRT_CHILD:
		return cr_rstrt_child(file);

	default:
		CR_KTRACE_BADPARM("unknown op %x", _IOC_NR(op));
		return -ENOTTY;
	}
}

static int ctrl_open(struct inode * inode, struct file * file)
{
	cr_pdata_t *priv;

	CR_KTRACE_FUNC_ENTRY();

	priv = kmalloc(sizeof(*priv), GFP_KERNEL);
	if (priv) {
		memset(priv, 0, sizeof(*priv));
		file->private_data = (void*)priv;
	}

	return priv ? 0 : -ENOMEM;
}

static int ctrl_release(struct inode * inode, struct file * file)
{
	cr_pdata_t *priv;

	CR_KTRACE_FUNC_ENTRY();

	priv = file->private_data;
	if (priv) {
		cr_chkpt_req_release(file, priv);
		// cr_chkpt_opt_release(file, priv);
		cr_rstrt_req_release(file, priv);
		cr_phase1_release(file, priv);
		cr_phase2_release(file, priv);
		kfree(priv);
	}

	return 0;
}

/*
 * Ops for the C/R control node
 */
struct file_operations cr_ctrl_fops =
{
	owner:		THIS_MODULE,
	open:		ctrl_open,
	release:	ctrl_release,
	ioctl:		ctrl_ioctl,
};
