/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_rstrt_req.c,v 1.45 2003/11/14 19:55:14 phargrov Exp $
 */

#include <linux/module.h>
#include <linux/errno.h>

#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/time.h>
#include <linux/mman.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>

#define __KERNEL_SYSCALLS__ 
#include <linux/unistd.h>

#include "cr.h"
#include "cr_module.h"
#include "cr_context.h"

#include "vmadump.h"

static int errno;

/* Fork with the given flags, stack and regs
 * Returns the pid of the new task, or negative error code.
 * At present we don't need the (soon to be obliterated) pid.
 */
static int
cr_do_fork(unsigned long clone_flags,
           unsigned long stack_start,
           struct pt_regs *regs)
{
#if HAVE_2_4_DO_FORK
    return do_fork(clone_flags, stack_start, regs, 0);
#elif HAVE_RH9_DO_FORK
    struct task_struct *task = do_fork(clone_flags, stack_start, regs, 0, NULL, NULL);
    return IS_ERR(task) ? PTR_ERR(task) : task->pid;
#else
    #error "no do_fork() ?"
#endif
}

/*
 * alloc_rstrt_request
 */
static cr_rstrt_req_t *
alloc_rstrt_req(void)
{
    cr_rstrt_req_t *req;

    req = kmalloc(sizeof(*req), GFP_ATOMIC);
    if (req) {
        // MOD_INC_USE_COUNT;
        memset(req, 0, sizeof(*req));
        atomic_set(&req->ref_count, 1);
	req->linkage = NULL;
	req->cr_restart_task = current;
	get_task_struct(req->cr_restart_task);
        set_bit(0, &req->need_linkage);
        set_bit(0, &req->need_close_control_fd);
        set_bit(0, &req->need_fs);
        set_bit(0, &req->need_files);
        set_bit(0, &req->need_hide_control_fd);
        clear_bit(0, &req->need_clones);
	init_MUTEX(&req->mutex);
	init_waitqueue_head(&req->wait);
	req->result = 0;
    }

    return req;

}

static void
release_rstrt_req(cr_rstrt_req_t *req)
{
    CR_KTRACE_REFCNT("ref count is approximately %d", 
		     atomic_read(&req->ref_count));

    if (atomic_dec_and_test(&req->ref_count)) {
        CR_KTRACE_REFCNT("fput()'ing file");
	fput(req->file);
	if (req->linkage) {
	    CR_KTRACE_ALLOC("kfree()'ing linkage");
	    kfree(req->linkage);
	}
	put_task_struct(req->cr_restart_task);
	CR_KTRACE_ALLOC("kfree()'ing req");
        kfree(req);
        // MOD_DEC_USE_COUNT;
    } else {
	wake_up(&req->wait);
    }
}

static int
make_task_req(void)
{
    return -ENOSYS;
}

static int
make_proc_req(void)
{
    return -ENOSYS;
}

#if CR_FILES_WORK
/*
 * source get typed.
 *
 * Doesn't quite get concurrency.
 */
static struct file *
__cr_src_gett(cr_rstrt_req_t *req, cr_obj_t object)
{
    return req->file;
}

/*
 * cr_src_gett
 *
 * source get typed.  Knows what we're planning to write, and understands
 * how to serialize when necessary.
 */
static struct file *
cr_src_gett(cr_rstrt_req_t *req, cr_obj_t object)
{
    return __cr_src_gett(req, object);
}

/*
 * __cr_src_putt
 *
 * stands for src put typed.
 */
static int
__cr_src_putt(struct file *filp)
{
    return 0;
}

/*
 * cr_src_putt
 *
 * stands for src put typed.
 */
static int
cr_src_putt(struct file *filp)
{
    return __cr_src_putt(filp);
}
#endif /* CR_FILES_WORK */

static int
cr_linkage_make_restart_task(struct task_struct *cr_restart_task,
                             struct cr_context_tasklinkage *l)
{
    /* We don't want any myptr's to match this by accident... */
    l->myptr = NULL;
    l->real_parent = NULL;
    l->parent = NULL;

#if HAVE_2_4_TASK_LINKAGE
    INIT_LIST_HEAD(&l->thread_group);
    /* Probably won't need that either */
#endif

    l->pid = cr_restart_task->pid;
    l->pgrp = cr_restart_task->pgrp;
    l->tty_old_pgrp = cr_restart_task->tty_old_pgrp;
    l->session = cr_restart_task->session;
    l->tgid = cr_restart_task->tgid;
    l->leader = cr_restart_task->tgid;
    /* 
     * Or any of that stuff
     */

    l->task = cr_restart_task;
    /*
     * But we need that one
     */

    return 0;
}

static int
cr_linkage_make_init_task(struct task_struct *task,
                          struct cr_context_tasklinkage *l)
{

    if (task == NULL) {
        CR_ERR("%s [%d]:  ERROR:  No task found!", __FUNCTION__, current->pid);
        return -1;
    }

    if (l == NULL) {
        CR_ERR("%s [%d]:  ERROR:  NULL linkage!", __FUNCTION__, current->pid);
        return -EINVAL;
    }

    l->myptr = CR_PARENT_INIT;
    l->task  = task;
    l->pid   = 1;

    /* XXX: not clear to me (PHH) that anything below this point is needed AT ALL */

    /*
     * This will make sure our links work correctly...
     * XXX:  But we need to make sure that appending the lists together
     * is handled specially at the end...
     */
    l->real_parent = NULL;
    l->parent = NULL;

#if HAVE_2_4_TASK_LINKAGE
    INIT_LIST_HEAD(&l->thread_group);
#endif

    l->pgrp = task->pgrp;
    l->tty_old_pgrp = task->tty_old_pgrp;
    l->session = task->session;
    l->tgid = task->tgid;
    l->leader = task->tgid;

    return 0;
}

/*
 * cr_load_linkage
 *
 * Loads the linkage structure stored in the context file into a new
 * buffer attatched to request
 *
 * This linkage structure is read, later, by cr_restore_linkage
 *
 * This grabs and releases tasklist_lock.
 *
 * NOTE:  Can use this later to reconstruct the graph of processes,
 * to restore the session and thread counts
 */
static int cr_load_linkage(cr_rstrt_req_t *req)
{
    struct cr_context_tasklinkage *linkage;
    int linkage_size;
    int retval;

    /* find out how large the linkage structure is */
    retval = cr_kread(req->file, &linkage_size, sizeof(linkage_size));
    if (retval < sizeof(linkage_size)) {
        goto out;
    }

    /* allocate memory for the linkage structure + 2 extras */
    linkage = NULL;
    if (linkage_size + 2*sizeof(*linkage) < 131072) {
        linkage = kmalloc(linkage_size + 2*sizeof(*linkage), GFP_KERNEL);
    }
    if (linkage == NULL) {
        CR_WARN("Too many processes (size = %d)", linkage_size);
        retval = -ENOMEM;
	goto out;
    }

    /* Now load the linkage from disk */
    /* NOTE:  The first two entries are used later for cr_restart and
     * init */
    retval = cr_kread(req->file, &linkage[2], linkage_size);
    if (retval < linkage_size) {
        kfree(linkage);
        goto out;
    }

    /* Add linkage entry for cr_restart task */
    cr_linkage_make_restart_task(req->cr_restart_task, &linkage[0]);
    /* Add linkage entry for init */
    read_lock(&tasklist_lock);
    cr_linkage_make_init_task(find_task_by_pid(1), &linkage[1]);
    read_unlock(&tasklist_lock);

    /* attatch the linkage structure to the request */
    req->linkage = linkage;
    req->linkage_size = linkage_size/sizeof(*linkage) + 2;
    retval = 0;

out:
    return retval;
}

/*
 * fork_current
 *
 * Creates a new process.
 */ 
static int
fork_current(void)
{
    int retval;
    struct pt_regs *regs;

    // ... Create the new process
    regs = get_pt_regs(current);

    retval = cr_do_fork(SIGCHLD, get_stack_pointer(regs), regs);

    return retval;
}

#if HAVE_RH9_TASK_LINKAGE
/* Code based heavily on kernel/pid.c */
#define BITS_PER_PAGE (8*PAGE_SIZE)
#define BITS_PER_PAGE_MASK (BITS_PER_PAGE-1)
typedef struct pidmap {
        atomic_t nr_free;
        void *page;
} pidmap_t;
#if CR_KSYM_pidmap_array
  extern pidmap_t pidmap_array[];
  #define pidmap_array CR_IMPORT_KSYM(pidmap_array)
#endif
#if CR_KSYM_pidmap_lock
  extern spinlock_t pidmap_lock;
  #define pidmap_lock CR_IMPORT_KSYM(pidmap_lock)
#endif
static inline void
_cr_free_pid(int pid)
{
    pidmap_t *map = &pidmap_array[pid / BITS_PER_PAGE];
    int offset = pid & BITS_PER_PAGE_MASK;

    clear_bit(offset, map->page);
    atomic_inc(&map->nr_free);
}
static int
_cr_alloc_pid(int pid) /* MUST call with no locks held */
{
    pidmap_t *map = &pidmap_array[pid / BITS_PER_PAGE];
    int offset = (pid & BITS_PER_PAGE_MASK);
    int result = 0;	/* assume failure */

    /* make sure we have a page */
    if (unlikely(map->page == NULL)) {
	unsigned long page = get_zeroed_page(GFP_KERNEL);

	spin_lock(&pidmap_lock);
	if (map->page) {
	    /* lost a race, free our page */
	    free_page(page);
	} else {
	    map->page = (void *)page;
	}
	spin_unlock(&pidmap_lock);
    }
    /* XXX assert (map->page != NULL) */

    if (!test_and_set_bit(offset, map->page)) {
	atomic_dec(&map->nr_free);
	result = 1;
    }

    return result;
}

/* Reserve all the pids we need
 * Returns non-zero on failure
 */
static int
cr_reserve_ids(cr_rstrt_req_t *req)
{
    struct cr_context_tasklinkage *link;
    int i;
    
    CR_KTRACE_HIGH_LVL("Now reserving required ids... ");

    /* Don't need to worry about session IDs, since exactly one member
     * of the session, the leader, will have a matching PID.  (When that
     * leader exits the entire session dies).
     */

#if 0 /* Not yet */
    /* _DO_ need to think about PGIDs.  Since it is possible for the process
     * which originally had pid==pgid to exit w/o terminating the process group,
     * we must call _cr_alloc_pid() for any such ids.  Easy way is to do this
     * for every pgid we restore and filter the pid==pgid cases in the pid loop.
     */
    foreach_unique_pgid_req {
        if (!_cr_alloc_pid(pgid)) {
            goto out_release_pgids;
        }
    }
#endif
    
    for (i = 2; i < req->linkage_size; ++i) {
	link = &req->linkage[i];

#if 0 /* Not yet */
	if (link->pid == link->pgid) {
	    /* we already grabbed this id in the pgid loop above */
	    continue;
	}
#endif

	if (!_cr_alloc_pid(link->pid)) {
#if CR_KERNEL_TRACING
	    read_lock(&tasklist_lock);
	    printk("Found pid %d in use: pid=%p, tgid=%p, pgid=%p, sid=%p\n",
		   link->pid,
		   find_pid(PIDTYPE_PID, link->pid),
		   find_pid(PIDTYPE_TGID, link->pid),
		   find_pid(PIDTYPE_PGID, link->pid),
		   find_pid(PIDTYPE_SID, link->pid));
	    read_unlock(&tasklist_lock);
#endif
	    goto out_release_pids;
	}
    }

    return 0;	/* success */

out_release_pids:
    /* If we failed w/ i=N, then 2<=j<N were allocated */
    while (i>2) {
	--i;
	_cr_free_pid(req->linkage[i].pid);
    }
#if 0	/* Not yet */
out_release_pgids:
    XXX release whichever pgids we had obtained before failure.
#endif

    return 1;	/* failure */
}

static void
cr_release_ids(cr_rstrt_req_t *req)
{
    struct cr_context_tasklinkage *link;
    int i;
    
    for (i = 2; i < req->linkage_size; ++i) {
	link = &req->linkage[i];
#if 0 /* Not yet */
	if (link->pid == link->pgid) {
	    /* we will free this id in the pgid loop below */
	    continue;
	}
#endif
	_cr_free_pid(link->pid);
    }

#if 0	/* Not yet */
    foreach_unique_pgid_req {
	_cr_free_pid(pgid);
    }
#endif
}
#endif /* HAVE_RH9_TASK_LINKAGE */

/*
 * cr_is_linkage_restorable
 *
 * Scans through a linkage and checks to see if any of the PIDs, etc.
 * that are described here are in use.  (i.e. check for PID conflict)
 * Also check for missing (NULL) linkage fields.
 *
 * NOTE:  Called from request_restart _AND_ from cr_do_restore_linkage!
 * As such, this thing _might_ be called with a spinlock held, so be
 * careful with what you go around doing in here, as you may get in trouble.
 *
 * NOTE: cr_is_linkage_restorable caller must read_lock tasklist_lock
 */
static int
cr_is_linkage_restorable(cr_rstrt_req_t *req)
{
    int retval;
    int i;
    struct cr_context_tasklinkage *l;
#if HAVE_2_4_TASK_LINKAGE
    struct task_struct *p;
#endif

    retval = 0;

    /* Make sure that cr_restart_task is around and initialized properly */
    if (req->linkage[0].task == NULL) {
        CR_ERR("%s [%d]:  NULL cr_restart_task!", __FUNCTION__, current->pid);
	return 0;
    }

    /* Try to avoid a possible race against req->cr_restart_task exiting
     * before we get to restore the linkage */
    if (req->linkage[0].task->state == TASK_ZOMBIE) {
        CR_ERR("%s [%d]:  Unrestorable linkage: cr_restart_task exited "
		"prematurely!", __FUNCTION__, current->pid);
        return 0;
    }

    for(i=2; i<req->linkage_size; ++i) {
	l = &req->linkage[i];
	
	if (!l->parent || !l->real_parent) {
	    CR_WARN("parent(s) not found ... i=%d", i);
            return 0;
	}
#if HAVE_2_4_TASK_LINKAGE
        if (l->pid >= PID_MAX) {
	    CR_WARN("PID out of bounds ... i=%d pid=%d", i, l->pid);
            return 0;
	}
        for_each_task(p) {
            if (l->pid == p->pid || l->tgid == p->tgid) {
		/* XXX: sessions not supported yet...
	        l->pgrp == p->pgrp || 
		l->sesion == p->session
		*/
	        CR_WARN("PID conflict ... i=%d pid=%d", i, l->pid);
                return 0;
	    }
	}
#else
	// XXX: should he validate some equivalent of PID_MAX ?
	if (!l->group_leader) {
	    CR_WARN("thread group leader not found ... i=%d", i);
            return 0;
	}
	if (find_pid(PIDTYPE_PID, l->pid) ||
	    find_pid(PIDTYPE_TGID, l->tgid)
	    /* XXX: sessions not supported yet...
	    find_pid(PIDTYPE_PGID, l->pgrp) ||
	    find_pid(PIDTYPE_SID, l->session) */
	   ) {
	    CR_WARN("PID conflict ... i=%d pid=%d", i, l->pid);
            return 0;
	}
#endif
    }

    return 1;
}

// cr_rstrt_req(user_req)
//
// Processes a restart request received from user space.
//
// Crazy return signature:
//   Never returns to child processes
//   Returns PID of child to parent process
//   Returns negative on failure
int cr_rstrt_request_restart(struct file *filp, void *arg)
{
    struct cr_context_header cf_header;
    cr_pdata_t *priv;
    struct cr_rstrt_args args;
    cr_rstrt_req_t *req;
    int retval;

    CR_KTRACE_FUNC_ENTRY();

    /* Make sure that we don't have any requests already on this fd */
    priv = filp->private_data;
    if (priv->rstrt_req || priv->chkpt_req) {
        return -EAGAIN;
    }

    /* XXX:  We don't handle a cr_restart that has children
     * yet, so make sure that you don't have any before continuing */
    if (CR_CHILD(current) != NULL) {
        CR_KTRACE_UNEXPECTED("Aborting.  Process has children.");
        return -EAGAIN;
    }

    /* Allocate a request structure for bookkeeping */
    req = alloc_rstrt_req();
    if (req == NULL) {
        return -ENOMEM;
    }
    priv->rstrt_req = req;
    
    /* Copy in the user's arguments */
    if (copy_from_user(&args, arg, sizeof(args))) {
	retval = -EFAULT;
        goto out_free_req;
    }

    // XXX:  Copy in user request options
    
    /* Check the file descriptor */
    // XXX:  Needs to be replaced with cr_rstrt_src_get()...
    req->file = fget(args.cr_fd);
    if (!(req->file)) {
	retval = -EBADF;
        goto out_free_req;
    }

    /* Read in the context file header */
    retval = cr_kread(req->file, &cf_header, sizeof(cf_header));
    if (retval != sizeof(cf_header)) goto out_free_req;

    CR_KTRACE_HIGH_LVL("cr_magic = %d %d, " 
		       "cr_version = %d, "
		       "checkpoint_type = %d, "
		       "num_threads = %d",
		       cf_header.magic[0], cf_header.magic[1], cf_header.version, 
		       cf_header.checkpoint_type, cf_header.num_threads);


    // XXX: Check cr_magic
    //
    if (cf_header.num_threads <= 0) {
        retval = -EBADSLT;
	goto out_free_req;
    }
    
    // Check cr_version
    if (cf_header.version != CR_CONTEXT_VERSION) {
        retval = -EBADRQC;
	goto out_free_req;
    }
    
    /* XXX: Have this check to make sure the type matches what the user
     * asked for. */
    switch(cf_header.checkpoint_type) {
        case CR_SCOPE_TASK:
	    retval = make_task_req();
	    break;
	case CR_SCOPE_PROC:
	    retval = make_proc_req();
	    break;
	case CR_SCOPE_PGRP:
	    retval = -ENOSYS;
	    goto out_free_req;
	    break;
	case CR_SCOPE_SESS:
	    retval = -ENOSYS;
	    goto out_free_req;
	    break;
	default:
	    retval = -EBADSLT;
	    goto out_free_req;
	    break;
    }
    
    /* Set number of threads for this one. */
    req->total_threads = cf_header.num_threads;
    atomic_set(&req->new_clones, cf_header.num_threads);
    clear_bit(0, &req->need_clones);
    if (cf_header.num_threads > 1) {
        /* Watch it with new_clones, if initialized to num_threads - 1,
	 * the atomic_dec_and_test loop will be off by one */
        set_bit(0, &req->need_clones);
    }

    /* Initialize parts of req relevant to restoring task linkage */
    /* First load the linkage */
    retval = cr_load_linkage(req);
    if (retval < 0) {
        CR_WARN("Error reading linkage structure! err=%d", retval);
	retval = -ENOMEM;
	goto out_free_req;
    }

#if HAVE_RH9_SIGNAL_STRUCT
    /* Check if signal struct is shared or not */
    /* XXX Need to read in real shared-pending signals here */
    retval = cr_kread(req->file, &req->shared_signal,
		      sizeof(req->shared_signal));
    if (retval < sizeof(req->shared_signal)) {
	retval = -EIO;
        goto out_free_req;
    }
#endif

    /* Now initialize the barrier count so we can block later */
    /* XXX:  This may not be correct.  We're assuming here that all the
     * threads that we create will actually call into the barrier code.
     * While this is true in all sane cases, it will not be true if one
     * of the threads dies prematurely...  pthreads has a habit of doing
     * this.  This should be fixed to one of the fail-safe barriers
     * defined in cr_barrier.h so we can abort properly.  No problem for
     * now, since we'll just time out if that ever happens */
    atomic_set(&req->link_barrier_count, cf_header.num_threads);
    init_waitqueue_head(&req->link_barrier_queue);

#if HAVE_RH9_TASK_LINKAGE
    /* Now we reserve all the ids (pid, pgid, sid) we will need.
     * Note that this can get_free_page(), and thus no locks are held yet.
     */
    if (cr_reserve_ids(req)) {
        CR_WARN("%s [%d]:  PID conflict found by cr_reserve_ids()", __FUNCTION__, current->pid);
	retval = -EBUSY;
	goto out_free_req;	
    }
#endif

    /*
     * Check to see if the linkage is restorable
     */
    read_lock(&tasklist_lock);
    retval = cr_is_linkage_restorable(req);
    read_unlock(&tasklist_lock);
    if (!retval) {
        CR_WARN("%s [%d]:  PID conflict found by cr_is_linkage_restorable()", __FUNCTION__, current->pid);
        retval = -EBUSY;
	goto out_free_pids;
    }

    /* Spawn off a child process to continue the work */
    atomic_inc(&req->ref_count);
    retval = fork_current();
    if (retval < 0) {
        CR_ERR("Unable to create a child process!  err=%d", retval);
	atomic_dec(&req->ref_count);
	retval = -ENOMEM;
	goto out_free_pids;
    }

    /* Normal case.  We don't want to ix the req, since we will need it
     * again */
    return retval;

out_free_pids:
#if HAVE_RH9_TASK_LINKAGE
    cr_release_ids(req);
#endif
out_free_req:
    release_rstrt_req(req);
    priv->rstrt_req = NULL;

    return retval;
}

// Check that a request is "done", from the point of view of the
// requestor.
// Returns >0 is completed, 0 if not completed, <0 on error
static
int check_done(cr_rstrt_req_t *req)
{
	/* Check that it is just me here, no children */
	return (atomic_read(&req->ref_count) == 1) ? 1 : 0;
}
	
// cr_rstrt_done(timeval)
//
// Test/wait for a restart request to complete (tasks are alive).
// Argument is:
//    Pointer to a (struct timeval) for a bounded wait
//    	if zero then we take care not to schedule().
//    NULL for an unbounded wait
//
// Returns:
//	<0 on errors (including interrupted sleep)
//	>0 if complete
//	0 if not complete within the timeout (if any)
//
// XXX: *should* return PID of "root" process (session leader)
int cr_rstrt_done(struct file *filp, void *arg)
{
    cr_pdata_t		*priv;
    cr_rstrt_req_t	*req;
    struct timeval	timeval;
    DECLARE_WAITQUEUE(wait, current);
	
    unsigned long timeout;
    int retval = 0;

    CR_KTRACE_FUNC_ENTRY();

    priv = filp->private_data;
    if (!priv) {
	return -EINVAL;
    }
    req = priv->rstrt_req;
    if (!req) {
	return -EINVAL;
    }

    // Fetch/decode the optional timeout from userspace
    timeout = MAX_SCHEDULE_TIMEOUT;
    if (arg) {
	retval = copy_from_user(&timeval, arg, sizeof(timeval));
	if (retval) {
	    return retval;
	}

	timeout = timeval_to_jiffies(&timeval);
    } 

    add_wait_queue(&req->wait, &wait);
    for (;;) {
	set_current_state(TASK_INTERRUPTIBLE);

	retval = check_done(req);

	if (retval || !timeout || signal_pending(current)) {
	    break;
	}

	timeout = schedule_timeout(timeout);
    }
    current->state = TASK_RUNNING;
    remove_wait_queue(&req->wait, &wait);

    // Copy remaining time back to userspace if applicable
    if (arg) {
	jiffies_to_timeval(timeout, &timeval);
	copy_to_user(arg, &timeval, sizeof(timeval));
	// like select() we fail silently if user timeval is R/O
    }

    // Check if we were interrupted
    if (!retval && timeout) {
	retval = -EINTR;
    }

    return retval;
}

// cr_rstrt_reap()
//
// Reap a completed restart.
// Returns:
// 	req->result on success
// 	-EINVAL if no un-reaped restart request is associated w/ this fd
// 	-EAGAIN if the request is not completed
int
cr_rstrt_reap(struct file *filp)
{
    cr_pdata_t *priv;
    cr_rstrt_req_t *req;
    int retval;

    CR_KTRACE_FUNC_ENTRY();

    retval = -EINVAL;
    priv = filp->private_data;
    if (priv) {
	req = priv->rstrt_req;
	if (req) {
	    retval = -EAGAIN;
	    if (check_done(req)) {
		priv->rstrt_req = NULL;
		retval = req->result;
		(void) release_rstrt_req(req);
	    }
	}
    }

    return retval;
}

/*
 * clone_current
 *
 * Sort of does a fork.  Allocates a stack for a new thread, copies
 * the current stack to it, and creates a new thread executing in this
 * duplicate stack.
 *
 * XXX:  For some reason, pthreads doesn't like it when we do this.  So
 * make sure that current is not linked with pthreads, otherwise the threads
 * seem to evaporate.
 *
 * Returns PID of new thread
 */
static int 
clone_current(cr_rstrt_req_t *req)
{
    int retval;
    unsigned long new_stack_start;
    unsigned long off;
    unsigned long sp;
    unsigned long buf;
    unsigned long flags;
    struct pt_regs *regs;
    struct vm_area_struct *current_stack_vma;

    regs = get_pt_regs(current);
    sp = get_stack_pointer(regs);

    /* find the vma for the current stack pointer */
    down_read(&current->mm->mmap_sem);
    current_stack_vma = find_vma(current->mm, sp);
    up_read(&current->mm->mmap_sem);
    if (!current_stack_vma)
    {
        retval = -ENOMEM;
	goto out;
    }

    /* allocate a vma to hold the new thread stack */
    down_write(&current->mm->mmap_sem);

    new_stack_start = do_mmap(0, 0,
			      current_stack_vma->vm_end -
			      current_stack_vma->vm_start,
			      PROT_READ | PROT_WRITE | PROT_EXEC,
			      MAP_PRIVATE | MAP_ANONYMOUS, 0);
    up_write(&current->mm->mmap_sem);
    if (!new_stack_start) {
	retval = -ENOMEM;
	goto out;
    }

    /* copy the old stack to the new stack */
    buf = __get_free_page(GFP_KERNEL);
    if (!buf) {
	retval = -ENOMEM;
	goto out;
    }
    for (off = 0;
	 off < current_stack_vma->vm_end - current_stack_vma->vm_start;
	 off += PAGE_SIZE) {
	if (copy_from_user((void *) buf,
			   ((void *) (current_stack_vma->vm_start + off)),
			   PAGE_SIZE)) {
	    CR_WARN("%d: copy_from_user failed.  off=%lx", current->pid, off);
	    retval = -EFAULT;
	    goto err_copy;
	}
	if (copy_to_user((void *) (new_stack_start + off),
			 (void *) buf, PAGE_SIZE)) {
	    CR_WARN("%d: copy_to_user failed.  off=%lx", current->pid, off);
	    retval = -EFAULT;
	    free_page(buf);
	    goto err_copy;
	}
    }

    flags = CLONE_VM | CLONE_FS | CLONE_FILES | CLONE_SIGHAND;
#if HAVE_RH9_SIGNAL_STRUCT
    if (req->shared_signal) {
	flags |= CLONE_THREAD | CLONE_DETACHED;
    }
#endif

    /* Now let her rip! */
    retval = cr_do_fork(flags,
			new_stack_start + (sp - current_stack_vma->vm_start),
			regs);

err_copy:
    free_page(buf);
out:
    return retval;
}

/*
 * cr_linkage_find_by_pid
 *
 * Returns a pointer to a linkage entry with l.myptr == myptr
 * NULL if not found
 */
struct cr_context_tasklinkage *
cr_linkage_find_by_pid(cr_rstrt_req_t *req, 
                       int pid)
{
    int i;
    struct cr_context_tasklinkage *entry;

    entry = NULL;

    for (i=0; i<req->linkage_size && entry == NULL; ++i) {
        if (req->linkage[i].pid == pid) {
	    entry = &(req->linkage[i]);
	}
    }

    return entry;
}

/*
 * cr_linkage_find_by_old_task
 *
 * Returns a pointer to a linkage entry with l.myptr == myptr
 * NULL if not found
 */
struct cr_context_tasklinkage *
cr_linkage_find_by_old_task(cr_rstrt_req_t *req, 
                            struct task_struct *myptr)
{
    int i;
    struct cr_context_tasklinkage *entry;

    entry = NULL;

    if (myptr != NULL) {
        for (i=0; i<req->linkage_size && entry == NULL; ++i) {
            if (req->linkage[i].myptr == myptr) {
	        entry = &(req->linkage[i]);
	    }
        }
    }

    return entry;
}

/*
 * cr_linkage_find_by_new_task
 *
 * Returns a pointer to a linkage entry with l.task == task
 */
struct cr_context_tasklinkage *
cr_linkage_find_by_new_task(cr_rstrt_req_t *req,
                            struct task_struct *task)
{
    int i;
    struct cr_context_tasklinkage *entry;

    entry = NULL;

    for (i=0; i<req->linkage_size && entry == NULL; ++i) {
        if (req->linkage[i].task == task) {
	    entry = &req->linkage[i];
	}
    }

    return entry;
}

/*
 * cr_linkage_get_parent
 *
 * Figure out who your parent process ought to be
 *
 * This takes care of making sure that if your new parent is cr_restart, then
 * things are ok.  Also takes care of init.
 *
 * XXX:  Well, it will take care of init.  For right now, we just assume
 * that init_task is the cr_restart task.  so checkpointed tasks that used
 * to be children of init are now children of cr_restart.
 */
struct cr_context_tasklinkage *
cr_linkage_get_parent(cr_rstrt_req_t *req,
                      struct task_struct *old_pptr)
{
    struct cr_context_tasklinkage *parent_entry;

    parent_entry = cr_linkage_find_by_old_task(req, old_pptr);
    /* No entry for this process, we'll reattatch it as a new child
     * of cr_restart. */
    if (parent_entry == NULL) {
        parent_entry = &req->linkage[0];
    }

    /* XXX:  Kludge
     * We don't deal with children of init correctly yet.  This makes sure
     * that if your parent _WAS_ init at the time of checkpoint, that your
     * new parent will be cr_restart
     */
    if (parent_entry == &req->linkage[1]) {
        parent_entry = &req->linkage[0];
    }

    return parent_entry;
}

/*
 * cr_linkage_get_real_parent
 *
 * For now, just assume the same.  Can fix this later by just
 * duplicating the above code.
 */
struct cr_context_tasklinkage *
cr_linkage_get_real_parent(cr_rstrt_req_t *req,
                         struct task_struct *old_opptr)
{
    return cr_linkage_get_parent(req, old_opptr);
}

#if HAVE_RH9_TASK_LINKAGE
/*
 * Add task to all the linked lists and hashes
 *
 * Must be called w/ tasklist_lock held for writing
 */
static void
cr_do_hash(struct task_struct *task, int is_leader)
{
    SET_LINKS(task);
    attach_pid(task, PIDTYPE_PID, task->pid);
    if (is_leader) {
        attach_pid(task, PIDTYPE_TGID, task->tgid);
        /* XXX: not yet
        cr_attach_pid(task, PIDTYPE_PGID, task->pgid);
        cr_attach_pid(task, PIDTYPE_SID, task->session);
        */
    } else {
        link_pid(task, task->pids + PIDTYPE_TGID,
		 &task->group_leader->pids[PIDTYPE_TGID].pid);
    }
}

/*
 * Remove task from all the linked lists and hashes
 *
 * Must be called w/ tasklist_lock held for writing
 */
static void
cr_do_unhash(struct task_struct *task)
{
    detach_pid(task, PIDTYPE_PID);
    detach_pid(task, PIDTYPE_TGID);
    if (thread_group_leader(task)) {
	/* XXX: not yet
	detach_pid(task, PIDTYPE_PGID);
	detach_pid(task, PIDTYPE_SID);
	*/
    }
    REMOVE_LINKS(task);
}
#endif


/* 
 * cr_do_restore_linkage
 *
 * NOTE:  Must only be called by a single thread.  This function acts on
 * every process contained in the restart request.  All other threads
 * should be blocked and prevented from returning to user space before entry
 * into this function.
 *
 * NOTE:  Must be called with tasklist_lock held.
 * Watch it inside of here, since you're holding a spinlock, you can't
 * do very much, and you can't waste too much time.  Just be concise and
 * to the point.
 *
 */
static int
cr_do_restore_linkage(cr_rstrt_req_t *req)
{
    int retval = 0;
    int i;
    struct cr_context_tasklinkage *link;

    /* XXX:  thread_group has to be handled specially, since we may only
     * construct this structure by fiddling with the task_struct directly.
     *
     * XXX:  thread_group is not restored
     */

    /*
     * Make a pass through to initialize some fields and to
     * translate old task pointers to new ones
     */
    for (i=0; i<req->linkage_size; ++i) {
        link = &req->linkage[i];
	if (link->task == NULL) {
            /* Means that we have an entry for which there is no
	     * corresponding task! */

	    /* XXX:  This can actually happen in normal cases if a thread
	     * exits...  Ignore for now...  Can probably pull this code
	     * out later. */
	    CR_ERR("%s [%d]:  ERROR:  No task for this PID. %d",
		     __FUNCTION__, current->pid, link->pid);
	    retval = -EINVAL;
	    goto out;
	}

	if (i < 2) {
	    /* Special case cr_restast task and init */
	    continue;
        }

	/* translate the linkage pointers from saved to restored */
	link->real_parent = cr_linkage_get_real_parent(req, link->real_parent)->task;
        link->parent = cr_linkage_get_parent(req, link->parent)->task;
#if HAVE_2_4_TASK_LINKAGE
	INIT_LIST_HEAD(&link->thread_group);
#elif HAVE_RH9_TASK_LINKAGE
	link->group_leader = cr_linkage_find_by_old_task(req, link->group_leader)->task;
#endif
    }

    /* Unlike the last pass, this pass is decisive.  If it doesn't work here,
     * then we _have_ to abort!  (or skip) */
    if (!cr_is_linkage_restorable(req)) {
        retval = -EBUSY;
        goto out;	
    }

    /*
     * init is special cased.  For now, just hack the data structure
     * so that references to the old init become references to the
     * cr_restart task
     */

    /* OK, now that the pointers have been calculated, we can go through
     * and actually perform the swap */
#if HAVE_2_4_TASK_LINKAGE
    for (i=2; i<req->linkage_size; ++i) {
	struct task_struct *task;

        link = &req->linkage[i];
	task = link->task;

	/* Remove existing links with parent, siblings, hashes, etc. */
	unhash_pid(task);
	REMOVE_LINKS(task);

        /* Set simple links to parent(s), aunts, uncles, etc. */
        task->p_pptr  = link->parent;
	task->p_opptr = link->real_parent;

	/* next set the various ids */
	task->tgid = link->tgid;
    	task->pid = link->pid;
	/* XXX: NOT YET
	task->pgrp = link->pgrp;
	task->session = link->session;
	*/

	task->exit_signal = link->exit_signal;

	/* Now relink, rehash for the new relatives and ids */
	SET_LINKS(task);
	hash_pid(task);

	// XXX:  Do thread_next

	/* XXX:  We don't handle these correctly yet */
#if 0
	task->tty_old_pgrp = link->tty_old_pgrp;
	task->leader = leader;
#endif
    }
#elif HAVE_RH9_TASK_LINKAGE
    /* Keeping the pid hashes/lists correct requires multiple passes */

    /* Pass 1: unhash and unlink all the tasks */
    for (i=2; i<req->linkage_size; ++i) {
	/* Remove existing links with parent, siblings, hashes, etc. */
	struct task_struct *task = req->linkage[i].task;
	cr_do_unhash(task);
    }

    /* Pass 2: change all the simple links and ids */
    for (i=2; i<req->linkage_size; ++i) {
	struct task_struct *task;

        link = &req->linkage[i];
	task = link->task;

        /* Set simple links to (step-)parent and task group leader */
        task->parent       = link->parent;
	task->real_parent  = link->real_parent;
	task->group_leader = link->group_leader;

	/* next set the various ids */
	task->tgid = link->tgid;
    	task->pid = link->pid;
	/* XXX: NOT YET
	task->pgrp = link->pgrp;
	task->session = link->session;
	*/

	task->clear_child_tid = link->clear_child_tid;
	task->exit_signal = link->exit_signal;

	/* XXX:  We don't handle these correctly yet */
#if 0
	task->tty_old_pgrp = link->tty_old_pgrp;
	task->leader = leader;
#endif
    }

    /* Pass 3: relink/rehash only thread group leaders
     * These must exist before their followers can be linked in. */
    for (i=2; i<req->linkage_size; ++i) {
	struct task_struct *task = req->linkage[i].task;

        if (thread_group_leader(task)) {
		cr_do_hash(task, 1);
	}
    }

    /* Pass 4: relink/rehash non-leaders */
    for (i=2; i<req->linkage_size; ++i) {
	struct task_struct *task = req->linkage[i].task;

        if (!thread_group_leader(task)) {
		cr_do_hash(task, 0);
	}
    }
#endif

    /* XXX A hack to return a "lead" pid. */
    for (i=2; i<req->linkage_size; ++i) {
	struct task_struct *task = req->linkage[i].task;

        if (CR_PARENT(task) == req->cr_restart_task) {
	    req->result = task->pid;
	}
    }

    /* on success retval is zero when we get here */

out:
    return retval;
}

/*
 * cr_restore_linkage
 *
 * This function implements a barrier, and hence must be called by all
 * restarting tasks before their return to user space.  Once the barrier
 * condition is reached (tasks entered == tasks created) the last thread
 * to enter restores the linkage.  At that point, the other tasks are
 * woken, and all tasks all allowed to continue
 *
 * This grabs tasklist_lock.
 *
 * Use cr_load_linkage to place the linkage structure into memory before
 * this function is called.
 *
 * Two _magic_ values in linkage structure:
 *   linkage[0] corresponds to the _original_ cr_restart task
 *     XXX:  Should this change to the initial baby?  -- sessions?
 *   linkage[1] corresponds to the init task
 */
static int
cr_restore_linkage(cr_rstrt_req_t *req, int pid)
{
    int retval;
    int my_old_pid;
    struct cr_context_tasklinkage *my_entry;

    /* Save this for debugging */
    my_old_pid = current->pid;

    /* Associate your old PID with your task in the linkage */
    my_entry = cr_linkage_find_by_pid(req, pid);
    if (my_entry == NULL) {
        CR_ERR("%s [%d]: ERROR:  No linkage entry for PID %d",
	       __FUNCTION__, current->pid, pid);
	/* XXX FIXME XXX FIXME XXX FIXME XXX FIXME XXX FIXME XXX */
	/* XXX                                               XXX */
	/* XXX:  THIS IS ABSOLUTELY THE WRONG THING TO DO!!! XXX */
	/* XXX                                               XXX */
	/* XXX FIXME XXX FIXME XXX FIXME XXX FIXME XXX FIXME XXX */
        return -1;
    }
    my_entry->task = current;

    /* First, we need to stop, so that none of the children start to
     * execute until their PIDs have been restored. */

    if (atomic_dec_and_test(&req->link_barrier_count)) {
        /* True only if we're the last one to enter the barrier, i.e. all
	 * tasks have successfully blocked and are waiting for us to do
	 * the remaining work. */

        CR_KTRACE_HIGH_LVL("Now restoring the linkage... ");
        /* Now grab tasklist_lock while we futz with everyones pointers */
        write_lock(&tasklist_lock);
	
	/* Do it! */
        cr_do_restore_linkage(req);

	/* Let the rest of the system come back */
	write_unlock(&tasklist_lock);

        CR_KTRACE_HIGH_LVL("Linkage restore finished...");

	/* At this point we're finished, so we can allow everyone else to
	 * complete their restart */
	/* Note: plain "clear_bit()" is not atomic */
	(void)test_and_clear_bit(0, &req->need_linkage);
	wake_up(&req->link_barrier_queue);

	retval = 0;
    } else {
        /* Wait for our PID and linkage stuff to be restored.  Note that at
	 * this point our PID will have changed.  */
	/* XXX: Should this be an interruptible wait?
	   This is a rather bad place to get broken out of the wait by ^C and
	   start running w/ only partial linkage restored!
	   On the other hand, we need a way to wake up in case of another
	   thread exiting unexpectedly on us. */
	wait_event(req->link_barrier_queue, !test_bit(0, &req->need_linkage));

        /* If any work remains to be done wrt linkage restore on a
	 * per-task basis, it should be done here.  */
	/* Nothing to do, so just return successful */
	retval = 0;  
    }

    return retval;
}

#if CR_FILES_WORK
/*
 * cr_fgets
 *
 * Sort of like fgets.
 *   Returns >= 0 on success.
 *   0 return means string is NULL.
 *   Returns number of bytes read in bytesread.
 */
int
cr_fgets(char *buf, int size, struct file *filp, int *bytesread)
{
    int ret;
    int read;
    int len = 0;

    ret = -EINVAL;
    if (!(bytesread)) {
        CR_ERR("cr_fgets: NULL bytesread passed as argument!");
        goto out;        
    }
    read = 0;

    ret = -EINVAL;
    if (!(buf)) {
        CR_ERR("cr_fgets: NULL buffer pass as argument!");
        goto out_bytesread;
    }

    ret = -EINVAL;
    if ((size < 0) || (size > PAGE_SIZE)) {
        CR_ERR("cr_fgets:  Bad buffer size.");
        goto out;
    }

    ret = cr_kread(filp, &len, sizeof(len));
    if (ret < 0) {
        CR_ERR("cr_fgets: Bad file read.");
        goto out_bytesread;
    } else if (ret != sizeof(len)) {
        read = ret;
        ret = -EIO;
        CR_ERR("cr_fgets: Short read.");
        goto out_bytesread;
    }
    read = ret;

    if (len == 0) {
        /* old string was NULL.  i.e. never initialized rather than a
	 * pointer to an empty string */
        ret = 0;
        goto out_bytesread;
    } else if ((len < 0) || (len > PAGE_SIZE) || (len > size)) {
        ret = -EINVAL;
        CR_ERR("cr_fgets:  Bad string length %d in file.", len);
        goto out_bytesread;
    } 

    ret = cr_kread(filp, buf, len);
    if (ret < 0) {
        CR_ERR("cr_fgets:  Bad file read.");
        goto out;
    } else if (ret != len) {
        read += ret;
        ret = -EIO;
        CR_ERR("cr_fgets:  Short read.");
        goto out_bytesread;
    }
    read += ret;

out_bytesread:
    *bytesread = read;
out:
    return ret;
}

/*
 * From Linux 2.4.20 fs/file.c
 */
fd_set * cr_alloc_fdset(int num)
{
        fd_set *new_fdset;
        int size = num / 8;

        if (size <= PAGE_SIZE)
                new_fdset = (fd_set *) kmalloc(size, GFP_KERNEL);
        else
                new_fdset = (fd_set *) vmalloc(size);
        return new_fdset;
}

/*
 * From Linux 2.4.20 fs/file.c
 */
void cr_free_fdset(fd_set *array, int num)
{
        int size = num / 8;

        if (!array) {
                printk(KERN_ERR "%s array = 0 (num = %d)\n",
                                __FUNCTION__, num);
                return;
        }

        if (num <= __FD_SETSIZE) /* Don't free an embedded fdset */
                return;
        else if (size <= PAGE_SIZE)
                kfree(array);
        else
                vfree(array);
}

/*
 * cr_restore_files_max_fdset
 *
 * Restores our fd_array and fdset_array to correct values to reopen old files.
 *
 * Adapted from expand_fdset in fs/file.c from Linux 2.4.20
 */
static int
cr_restore_files_max_fdset(struct files_struct *files, int nr)
{
    fd_set *new_cloexec_fds = NULL;
    fd_set *new_open_fds = NULL;
    int retval;

    /* check system limits */
    retval = -EMFILE;
    if (files->max_fds >= NR_OPEN || nr >= NR_OPEN)
        goto out;

    retval = 0;
    write_lock(&files->file_lock);
    if (nr <= files->max_fdset) {
	write_unlock(&files->file_lock);
	goto out;
    }

    retval = -ENOMEM;
    write_unlock(&files->file_lock);
    new_cloexec_fds = cr_alloc_fdset(nr);
    new_open_fds = cr_alloc_fdset(nr);
    if (!new_open_fds || !new_cloexec_fds) {
	CR_ERR("Unable to allocate fdset.");
	goto out;
    }

    retval = 0;
    write_lock(&files->file_lock);
    /* Copy the existing tables and install the new pointers */
    if (nr > files->max_fdset) {
	int i = files->max_fdset / (sizeof(unsigned long) * 8);
	int count = (nr - files->max_fdset) / 8;

	/* 
	 * Don't copy the entire array if the current fdset is not yet initialised.  
	 */
	if (i) {
	    memcpy(new_open_fds, files->open_fds, files->max_fdset / 8);
	    memcpy(new_cloexec_fds, files->close_on_exec, files->max_fdset / 8);
	    memset(&new_open_fds->fds_bits[i], 0, count);
	    memset(&new_cloexec_fds->fds_bits[i], 0, count);
	}

	nr = xchg(&files->max_fdset, nr);
	new_open_fds = xchg(&files->open_fds, new_open_fds);
	new_cloexec_fds = xchg(&files->close_on_exec, new_cloexec_fds);
	write_unlock(&files->file_lock);
	cr_free_fdset(new_open_fds, nr);
	cr_free_fdset(new_cloexec_fds, nr);
	return 0;
    }
    /* Somebody expanded the array while we slept ... */

    write_unlock(&files->file_lock);

out:
    if (new_open_fds)
	cr_free_fdset(new_open_fds, nr);
    if (new_cloexec_fds)
	cr_free_fdset(new_cloexec_fds, nr);
    return retval;
}


/*
 * frees an fd_array.
 *
 * Borrowed from Linux 2.4.20 fs/file.c
 */
void cr_free_fd_array(struct file **array, int num)
{
        int size = num * sizeof(struct file *);

        if (!array) {
                printk(KERN_ERR "%s array = 0 (num = %d)\n",
                                __FUNCTION__, num);
                return;
        }

        if (num <= NR_OPEN_DEFAULT) /* Don't free the embedded fd array! */
                return;
        else if (size <= PAGE_SIZE)
                kfree(array);
        else
                vfree(array);
}

/*
 * Restores the fd array to the proper size.
 *
 * Again, borrowed heavily from linux/fs/file.c
 */
static int
cr_restore_files_max_fds(struct files_struct *files, int nr)
{
    int size = nr * sizeof(struct file *);
    int retval;
    struct file **new_fd_array;

    write_lock(&files->file_lock);
    if (nr <= files->max_fds) {
	write_unlock(&files->file_lock);
	return 0;
    }

    retval = -ENOMEM;
    write_unlock(&files->file_lock);

    /* allocate new array */
    if (size <= PAGE_SIZE)
	new_fd_array = (struct file **) kmalloc(size, GFP_KERNEL);
    else
	new_fd_array = (struct file **) vmalloc(size);

    if (!new_fd_array)
	goto out;

    /* now copy the stuff in the old array into the new one. */
    write_lock(&files->file_lock);
    if (nr > files->max_fds) {
	struct file **old_fds;
	int i;

	old_fds = xchg(&files->fd, new_fd_array);
	i = xchg(&files->max_fds, nr);

	/* Don't copy/clear the array if we are creating a new fd array for fork() */
	if (i) {
	    memcpy(new_fd_array, old_fds, i * sizeof(struct file *));
	    /* clear the remainder of the array */
	    memset(&new_fd_array[i], 0, (nr - i) * sizeof(struct file *));

	    write_unlock(&files->file_lock);
	    cr_free_fd_array(old_fds, i);
	}
    } else {
	/* Somebody expanded the array while we slept ... */
	write_unlock(&files->file_lock);
	if (size <= PAGE_SIZE)
            kfree(new_fd_array);
        else
            vfree(new_fd_array);
    }
    retval = 0;

out:
    return retval;
}

/*
 * cr_restore_files_struct
 *
 * Restore the files_struct (file table).
 *
 * This function just expands our fd array to the correct size.
 *
 * In the future, if we need any information that applies to all opened files
 * that info can be placed here.
 */
static int
cr_restore_files_struct(cr_rstrt_req_t *req)
{
    struct cr_files_struct cr_fs;
    struct file *c_filp;
    int retval = 0;

    c_filp = cr_src_gett(req, cr_files_obj);
    if (!c_filp) {
	CR_ERR("Unable to find context file to load cr_files_struct.");
	goto out_no_cfilp;
    }

    retval = cr_kread(c_filp, &cr_fs, sizeof(cr_fs));
    if (retval != sizeof(cr_fs)) {
	CR_ERR("cr_fs: Bad file read!");
	retval = -EIO;
	goto out_putt;
    }

    retval = 0;
    if (cr_fs.cr_obj != cr_files_obj) {
	CR_ERR("cr_fs: Invalid cr_files_struct!");
	retval = -EINVAL;
	goto out_putt;
    }

    /* check to see if we have too many files open */
    retval = -EMFILE;
    if (((unsigned long) cr_fs.cr_max_fdset) > current->rlim[RLIMIT_NOFILE].rlim_cur) {
        CR_KTRACE_HIGH_LVL("Too many files %d > %lu", cr_fs.cr_max_fdset, current->rlim[RLIMIT_NOFILE].rlim_cur);
	goto out_putt;
    }

    retval = cr_restore_files_max_fdset(current->files, cr_fs.cr_max_fdset);
    if (retval < 0) {
	CR_ERR("Failed to resize the fdset array. (err=%d)", retval);
	goto out_putt;
    }

    retval = cr_restore_files_max_fds(current->files, cr_fs.cr_max_fds);
    if (retval < 0) {
	CR_ERR("Failed to resize the fd array. (err=%d)", retval);
	goto out_putt;
    }

    write_lock(&current->files->file_lock);
    if (cr_fs.cr_next_fd > current->files->max_fdset) {
        /* shouldn't ever happen */
        retval = -EINVAL;
        write_unlock(&current->files->file_lock);
	CR_ERR("Failed to restore next_fd.");
	goto out_putt;
    }

    current->files->next_fd = cr_fs.cr_next_fd;
    write_unlock(&current->files->file_lock);

    retval = 0;

out_putt:
    cr_src_putt(c_filp);
out_no_cfilp:
    return retval;
}

/*
 * cr_load_file_info
 * 
 * Reads in a file_info from disk, returns it.
 */
static int
cr_load_file_info(cr_rstrt_req_t *req, struct cr_file_info *file_info)
{
    struct file *filp;
    int retval;

    filp = cr_src_gett(req, cr_file_obj);
    if (!filp) {
        retval = -EBADF;
        goto out;
    }

    retval = cr_kread(filp, file_info, sizeof(*file_info));
    if (retval != sizeof(*file_info)) {
        CR_ERR("cr_load_file_info: Bad read! retval=%d", retval);
        retval = -EIO;
	goto out_putt;
    }

    retval = -EINVAL;
    if (file_info->cr_type != cr_file_info_obj) {
        CR_ERR("cr_load_file_info: Garbage in context file!");
	goto out_putt;
    }

    retval = 0;

out_putt:
    cr_src_putt(filp);
out:
    return retval;
}

/*
 * cr_open:  open a file - accepts file descriptor as argument
 * 
 * NOTE:  Does not attempt to expand the fdset array. i.e. aborts if fd > files->max_fdset
 * Use cr_restore_files_struct to expand fdset to the correct size.
 *
 * NOTE:  Does not check limits on number of open files.
 */ 
static int
cr_open(int fd, char *path, int flags, int mode)
{
    int retval;
    struct file *filp;

    /* check to see if we have too many files open */
    retval = -EMFILE;
    if (fd >= current->rlim[RLIMIT_NOFILE].rlim_cur)
	goto out;

    /* check to see if we have too many files open */
    retval = -EMFILE;
    if (fd >= current->files->max_fds)
	goto out;

    /* Mark the fd in use */
    write_unlock(current->files->file_lock);
    /* make sure it's not in use already.  shouldn't ever happen */
    if (FD_ISSET(fd, current->files->open_fds)) {
        CR_WARN("File opened unexpectedly.");
	retval = -EBUSY;
        write_unlock(current->files->file_lock);
        goto out;
    }
    FD_SET(fd, current->files->open_fds);
    FD_CLR(fd, current->files->close_on_exec);
    write_unlock(current->files->file_lock);

    /* open the file */
    filp = filp_open(path, flags, mode);
    if (IS_ERR(filp)) {
        retval = PTR_ERR(filp);
        goto out;
    }

    /* install the filp in the fd array -- note race see fd_install */
    fd_install(fd, filp);

    retval=0;

out:
    return retval;
}


/*
 * cr_restore_open_file
 *
 * reopens files
 */
static int
cr_restore_open_file(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval;
    char *buf;
    struct file *c_filp;
    struct cr_open_file open_file;
    int len;

    c_filp = cr_src_gett(req, cr_file_obj);
    if (!c_filp) {
        retval = -EBADF;
        goto out;
    }

    buf = (char *) __get_free_page(GFP_KERNEL);
    if (!buf) {
        CR_ERR("Couldn't allocate buffer for file name.");
        retval = -ENOMEM;
        goto out_putt;
    }
    
    /* read in the cr_open_file struct */
    retval = cr_kread(c_filp, &open_file, sizeof(open_file));
    if (retval != sizeof(open_file)) {
        CR_ERR("cr_restore_open_file - Bad read.");
        retval = -EIO;
        goto out_free;
    }

    if (open_file.cr_type != cr_open_file_obj) {
        CR_ERR("cr_restore_open_file.  Garbage in context file!");
    }

    /* now read out the name */
    retval = cr_fgets(buf, PAGE_SIZE, c_filp, &len);
    if (retval < 0) {
        CR_ERR("cr_restore_open_file - pwd Bad read!");
        retval = -EIO;
        goto out_free;
    } else if (retval == 0) {
        /* 0 corresponds to NULL, should never have happened. */
        CR_ERR("A file with no name?");
	retval = -EINVAL;
	goto out_free;
    }

    if (do_not_restore) {
        retval = 0;
	goto out_free;
    }

    /* reopen file */
    retval = cr_open(file_info->fd, buf, open_file.f_flags, open_file.i_mode);
    if(retval) {
        goto out_free;
    }

    /* restore position in file */
    retval = lseek(file_info->fd, open_file.f_pos, 0);
    if (retval < 0) {
        CR_ERR("Couldn't restore file pointer.");
	goto out_free;
    }

    retval = 0;

out_free:
    free_page((unsigned long) buf);
out_putt:
    cr_src_putt(c_filp);
out:
    return retval;
}

static int 
cr_restore_open_dir(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;

    if (do_not_restore) {
        retval = 0;
    }

    return retval;
}

static int 
cr_restore_open_link(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;

    if (do_not_restore) {
        retval = 0;
    }

    return retval;
}

static int 
cr_restore_open_fifo(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;
    struct cr_dup cf_dup;
    struct file *cf_filp;

    cf_filp = cr_src_gett(req, cr_chrdev_obj);
    if (!cf_filp) {
        retval = -EBADF;
        goto out;
    }

    /* read in the cr_chrdev struct */
    retval = cr_kread(cf_filp, &cf_dup, sizeof(cf_dup));
    if (retval != sizeof(cf_dup)) {
        CR_ERR("cr_restore_open_file - Bad read.");
        retval = -EIO;
        goto out_putt;
    }

    /* XXX:  Validate type */

    CR_KTRACE_LOW_LVL("   Open fifo: -1 == %d.", cf_dup.cr_dupfd);

    if (do_not_restore) {
        retval = 0;
	goto out_putt;
    }

    retval = 0;

out_putt:
    cr_src_putt(cf_filp);
out:
    return retval;
}

static int 
cr_restore_open_socket(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;

    if (do_not_restore) {
        retval = 0;
    }

    return retval;
}

static int 
cr_restore_open_chr(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;
    struct cr_chrdev cf_chrdev;
    struct file *c_filp;

    c_filp = cr_src_gett(req, cr_chrdev_obj);
    if (!c_filp) {
        retval = -EBADF;
        goto out;
    }

    /* read in the cr_chrdev struct */
    retval = cr_kread(c_filp, &cf_chrdev, sizeof(cf_chrdev));
    if (retval != sizeof(cf_chrdev)) {
        CR_ERR("cr_restore_open_file - Bad read.");
        retval = -EIO;
        goto out_putt;
    }

    /* XXX:  Validate type */
    retval = -EINVAL;
    if (cf_chrdev.cr_type != cr_open_chr) {
        CR_ERR("cr_restore_open_chr: Garbage in context file.");
        goto out_putt;
    }

    CR_KTRACE_LOW_LVL("   Character device.  major=%d, minor=%d", cf_chrdev.cr_major, cf_chrdev.cr_minor);

    if (do_not_restore) {
        retval = 0;
	goto out_putt;
    }

    retval = 0;

out_putt:
    cr_src_putt(c_filp);
out:
    return retval;
}

static int 
cr_restore_open_blk(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;

    if (do_not_restore) {
        retval = 0;
    }

    return retval;
}

static int 
cr_restore_open_dup(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;
    struct cr_dup cf_dup;
    struct file *cf_filp;

    cf_filp = cr_src_gett(req, cr_chrdev_obj);
    if (!cf_filp) {
        retval = -EBADF;
        goto out;
    }

    /* read in the cr_chrdev struct */
    retval = cr_kread(cf_filp, &cf_dup, sizeof(cf_dup));
    if (retval != sizeof(cf_dup)) {
        CR_ERR("cr_restore_open_file - Bad read.");
        retval = -EIO;
        goto out_putt;
    }

    /* XXX:  Validate type */
    retval = -EINVAL;
    if (cf_dup.cr_type != cr_open_dup) {
        CR_ERR("cr_restore_open_dup: Garbage in context file.");
        goto out_putt;
    }

    CR_KTRACE_LOW_LVL("   Dup of %d.", cf_dup.cr_dupfd);

    if (do_not_restore) {
        retval = 0;
	goto out_putt;
    }

    retval = 0;

out_putt:
    cr_src_putt(cf_filp);
out:
    return retval;
}

static int 
cr_end_of_file_restore(cr_rstrt_req_t *req, struct cr_file_info *file_info)
{
    int retval = 0;

    return retval;
}

static int
cr_restore_file_locks(cr_rstrt_req_t *req, struct cr_file_info *file_info)
{
    int retval = -ENOSYS;

    return retval;
}

static int
cr_finish_all_file_restores(cr_rstrt_req_t *req)
{
    int retval = -ENOSYS;

    return retval;
}

/*
 * cr_restore_all_files
 *
 * Main entry point for file recovery
 *
 * In the future, may want to set this so that it manipulates file "objects"
 * with load / restore methods
 */
static int
cr_restore_all_files(cr_rstrt_req_t *req)
{
    int restore_flag;
    int retval;
    struct cr_file_info file_info;


    /* load the files struct */
    CR_KTRACE_HIGH_LVL("recovering fs_struct...");
    retval = cr_restore_files_struct(req);
    if (retval) {
        CR_ERR("Unable to cr_restore_files_struct");
        goto out;
    }

    file_info.cr_file_type = cr_bad_file_type;
    /* now do per file stuff */
    do {
        retval = cr_load_file_info(req, &file_info);
	if (retval < 0) {
            CR_ERR("Error loading file_info.");
            goto out;
	}

	restore_flag = 0;
	read_lock(&current->files->file_lock);
	if (FD_ISSET(file_info.fd, current->files->open_fds)) {
            restore_flag = 1;
	}
	read_unlock(&current->files->file_lock);

        CR_KTRACE_LOW_LVL("   fd=%d flag=%d", file_info.fd, restore_flag);

        switch(file_info.cr_file_type) {
        case cr_open_file:
            retval = cr_restore_open_file(req, &file_info, restore_flag);
            break;
        case cr_open_directory:
            retval = cr_restore_open_dir(req, &file_info, restore_flag);
            break;
        case cr_open_link:
            retval = cr_restore_open_link(req, &file_info, restore_flag);
            break;
        case cr_open_fifo:
            retval = cr_restore_open_fifo(req, &file_info, restore_flag);
            break;
        case cr_open_socket:
            retval = cr_restore_open_socket(req, &file_info, restore_flag);
            break;
        case cr_open_chr:
            retval = cr_restore_open_chr(req, &file_info, restore_flag);
            break;
        case cr_open_blk:
            retval = cr_restore_open_blk(req, &file_info, restore_flag);
            break;
        case cr_open_dup:
            retval = cr_restore_open_dup(req, &file_info, restore_flag);
            break;
        case cr_end_of_files:
            /* finished */
            retval = cr_end_of_file_restore(req, &file_info);
            break;
        case cr_bad_file_type:
            /* fall through */
        default:
            CR_ERR("Unknown file type fd=%d type=%d", file_info.fd, file_info.cr_file_type);
            retval = -EBADF;
            goto out;
            break;
        }

        if (retval < 0) {
            CR_ERR("Unable to restore file.");
            goto out;
        }

        /* now acquire locks */
        cr_restore_file_locks(req, &file_info);

    } while (file_info.cr_file_type != cr_end_of_files);

    /* now for the second pass.  The above restore routines should have taken
     * care of loading all the data.  the easy work, now we do everything 
     * that has to be deferred until we have all the data
     */
    cr_finish_all_file_restores(req);

out:
    return retval;
} 

/*
 * cr_chdir
 *
 * Our own version of the chdir() system call.  Differs in that it accepts
 * pathnames in kernel space, and calls path_init, path_walk, and path_release
 * directly.
 * 
 */
static int
cr_chdir(const char *path)
{
    struct nameidata nd;
    int retval;

    /* lookup the path */
    retval = path_init(path,LOOKUP_POSITIVE|LOOKUP_FOLLOW|LOOKUP_DIRECTORY,&nd);
    if (retval)
        retval = path_walk(path, &nd);
    if (retval)
        goto out;

    retval = permission(nd.dentry->d_inode,MAY_EXEC);
    if (retval)
        goto dput_and_out;

    set_fs_pwd(current->fs, nd.mnt, nd.dentry);

dput_and_out:
    path_release(&nd);
out:
    return retval;
}

/*
 * cr_restore_fs
 *
 * This function does the work of restoring the data in our fs_struct, i.e.
 * the current working directory, and the umask.
 *
 * In the future, this may be modified to also restore a process after a
 * chroot(), if desirable.
 * 
 */
static int
cr_restore_fs_struct(cr_rstrt_req_t *req)
{
    int len;
    int retval;
    int umask;
    char *buf;
    struct file *filp;

    buf = (char *) __get_free_page(GFP_KERNEL);
    if (!buf) {
        retval = -ENOMEM;
        goto out_nomem;
    }

    filp = cr_src_gett(req, cr_fs_obj);
    if (!filp) {
        retval = -EBADF;
        goto out_nofile;
    }

    /* restore the umask */
    retval = cr_kread(filp, &umask, sizeof(umask));
    if (retval < sizeof(umask)) {
        CR_ERR("umask: Bad file read! retval=%d", retval);
        retval = -EIO;
        goto out_readerr;
    }

    /* read in root */
    retval = cr_fgets(buf, PAGE_SIZE, filp, &len);
    if (retval < 0) {
        CR_ERR("root: Bad file read!");
        goto out_readerr;
    }

    /* read in altroot */
    retval = cr_fgets(buf, PAGE_SIZE, filp, &len);
    if (retval < 0) {
        CR_ERR("altroot: Bad file read!");
        goto out_readerr;
    } else if (retval > 0) {
        CR_ERR("Warning:  chroot recovery not implemented.");
    }

    /* read in the current working directory */
    retval = cr_fgets(buf, PAGE_SIZE, filp, &len);
    if (retval < 0) {
        CR_ERR("pwd: Bad file read!");
        goto out_readerr;
    }

    /* should change root directory here. */
    /* cr_chroot() */

    /* now change the current working directory */
    retval = cr_chdir(buf);
    if (retval < 0) {
        CR_ERR("cr_chdir:  Unable to change directory to %s", buf);
	goto out_readerr;
    }    

out_readerr:
    cr_src_putt(filp);
out_nofile:
    free_page((unsigned long) buf);
out_nomem:
    return retval;
}
#endif /* CR_FILES_WORK */

// cr_rstrt_child
//
// Invoked by a freshly created child process.
//
// The real "restart" part of reconstructing a process image happens here.
//
// Returns 0 on success, or a negative on failure
int cr_rstrt_child(struct file *filp)
{
    cr_pdata_t *priv;
    cr_rstrt_req_t *req; 
    int retval;
    int old_pid;

    retval = -EINVAL;
    if (filp == NULL) {
        CR_ERR("%s: Called with NULL file structure", __FUNCTION__);
        goto out;
    }

    priv = filp->private_data;
    if (priv == NULL) {
        CR_ERR("%s: private_data is NULL!", __FUNCTION__);
        goto out;
    }

    req = priv->rstrt_req;
    if (req == NULL) {
        CR_ERR("%s: No rstrt_req attatched to filp!", __FUNCTION__);
        goto out;
    }

    /* Create clones of yourself if necessary */
    if (test_and_clear_bit(0, &req->need_clones)) {
        CR_KTRACE_HIGH_LVL("%d: Creating new processes", current->pid);
        while (!atomic_dec_and_test(&req->new_clones)) {
            atomic_inc(&req->ref_count);
	    retval = clone_current(req);
	    if (retval < 0) {
                CR_ERR("Unable to clone a child process!  err=%d", retval);
	        atomic_dec(&req->ref_count);
	        retval = -ENOMEM;
	        goto out1;
	    } else if (retval > 0) {
                CR_KTRACE_LOW_LVL("Clone returned %d to PID %d", 
				  retval, current->pid);
	    } else {
                /* This should never, EVER, happen. */
	        CR_ERR("cr_restart:  Incredibly confused.");
	    }
        }
    } else {
        CR_KTRACE_HIGH_LVL("%d: Have enough processes", current->pid);
    }

    /* Thaw thyself out */
    {
        struct vmadump_threadundump_args thaw_args;
	long err;

	thaw_args.num_threads = req->total_threads;
	thaw_args.fd = -1;              /* Ignored, but we initialize */
	thaw_args.flags = 0;            /* no flags */
	thaw_args.old_pid = -1;         /* Output, but initialize anyway */

        err = vmadump_thaw_threads(req->file, get_pt_regs(current), &thaw_args);
	if (err < 0) {
	    CR_ERR("thaw_threads returned error %ld", err);
	}
	retval = 0;

	/* save this so we can match ourselves up with the appropriate
	 * structure in the linkage later */
	old_pid = thaw_args.old_pid;
    }

    /* Restore linkage (PID, session ID, etc.) */
    CR_KTRACE_HIGH_LVL("Formerly %s PID %d", current->comm, old_pid);

    retval = cr_restore_linkage(req, old_pid);
    if (retval < 0) {
        CR_ERR("%s [%d]:  Unable to restore linkage!  (err=%d)",
	       __FUNCTION__, current->pid, retval);
        // XXX:  Get the hell out of here...
        // XXX:  goto ok here?
        goto out1;
    }

    /* --- File recovery starts here --- */
    /* Exactly one thread should close the rstrt_req file descriptor(s)
     * but they should all wait until this has been done. */
    /* NOTE:  Past this point, you must jump to out_close to abort.  We are removing
     * our reference to the fd, which means do_exit() can't clean up after us. */ 
    // XXX:  Does it matter which thread (first vs. last) does this?
    down(&req->mutex);
    if (test_and_clear_bit(0, &req->need_hide_control_fd)) {
        /* Remove the rstrt_req descriptor from our open files. */
        struct files_struct *files = current->files;
        int found;
        int fd;

	found = 0;
        write_lock(&files->file_lock);
	for (fd = 0; fd < files->max_fds; ++fd) {
            if (filp == files->fd[fd]) {
		found = 1;
		files->fd[fd] = NULL;
		FD_CLR(fd, files->close_on_exec);
		__put_unused_fd(files, fd);
	    }
	}
	write_unlock(&files->file_lock);

        if (!found) {
            CR_ERR("Checkpoint control descriptor closed unexpectedly!");
	}
    }
    up(&req->mutex);

#if CR_FILES_WORK
    /* restore fs_struct */
    /* XXX:  put the bit tests inside the restore routines when we do process
     * groups and sessions.  This way, they can just decide they want to skip it.
     */
    down(&req->mutex);
    if (test_and_clear_bit(0, &req->need_fs)) {
        retval = cr_restore_fs_struct(req);
        if (retval < 0) {
            CR_ERR("%s [%d]:  Unable to restore fs!  (err=%d)",
                   __FUNCTION__, current->pid, retval);

            up(&req->mutex);
            goto out_close;
            // XXX:  same concern
        }
    }
    up(&req->mutex);

    /* restore the open files */
    down(&req->mutex);
    if (test_and_clear_bit(0, &req->need_files)) {
        retval = cr_restore_all_files(req);
        if (retval < 0) {
            CR_ERR("%s [%d]:  Unable to restore files!  (err=%d)",
                   __FUNCTION__, current->pid, retval);
            up(&req->mutex);
            goto out_close;
            // XXX:  same concern
        }
    }
    up(&req->mutex);
#endif

    /*
     * can probably ease up on the use of all those mutex's...  We take care to
     * acquire the proper locks on the inside of the file restore routines,
     * but I'd rather test this later.
     */

    /* ok.  Now we can let everyone else proceed. */

    /* Mark the cr_task state so the restart is complete */

    /* close the request descriptor */
out_close:
    down(&req->mutex);
    if (test_and_clear_bit(0, &req->need_close_control_fd)) {
	int rc;	/* don't overwrite error code w/ 0 */
	CR_KTRACE_LOW_LVL("%d: closing request descriptor", current->pid);
	rc = filp_close(filp, current->files);
	if (rc < 0) {
	    CR_ERR("Error closing checkpoint control descriptor!  "
		   "(err=%d)", rc);
	    retval = rc;
	}
    }
    up(&req->mutex);
out1:
    /* XXX: this moves to rstrt_barrier once handler re-registration is done? */
    release_rstrt_req(req);
out:
    if (retval == 0) {
        /* Return > zero for success restart */
        retval = 1;
    }
    CR_KTRACE_FUNC_EXIT("Returning %d", retval);
    return retval;
}

// cr_rstrt_barrier
//
// Blocks until this process is permitted to continue
//
// Returns 0 on success, negative on failure
int cr_rstrt_barrier(struct file *filp, unsigned long arg)
{
    CR_ERR("cr_rstrt_barrier: Not implemented!");

    return -EINVAL;
}

int cr_rstrt_req_release(struct file *file, cr_pdata_t *priv)
{
    CR_KTRACE_FUNC_ENTRY();
    if (priv == NULL) {
        CR_WARN("Warning:  Called with NULL priv!");
        return -EINVAL;
    }

    if (priv->rstrt_req) {
        release_rstrt_req(priv->rstrt_req);
    }

    return 0;
}
