/* 
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_async.c,v 1.45 2003/11/14 19:55:24 phargrov Exp $
 *
 * This file deals with the checkpoint thread(s) run to execute callbacks
 * registered as CR_THREAD_CONTEXT.
 * The filename "cr_async.c" is legacy from when these callbacks were
 * known as "asynchronous handlers".
 */

#include <signal.h>	// for sigfillmask()
#include <errno.h>	// for errno

#include "cr_private.h"

#if LIBCR_TRACING
#	include <sys/types.h>	// for pid_t
#	include <unistd.h>	// for getpid()
#endif

//
// Private data
//

enum {
    CRI_THREAD_STOPPED,
    CRI_THREAD_STARTING,
    CRI_THREAD_RUNNING,
};

// Stuff for the callback thread, all protected by thread_lock:
static cri_info_t	*thread_info_p;
static pthread_mutex_t	thread_lock = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t	thread_init_cond = PTHREAD_COND_INITIALIZER;
static int		thread_state = CRI_THREAD_STOPPED;
static pthread_t	my_thread;

//
// Private functions
//

// my_handler()
//
// Signal handler for threads running thread-context callbacks
static void my_handler(int signr, siginfo_t *siginfo, void *context)
{
    int saved_errno = errno;
    cri_info_t *info = cri_info_location();	// thread-specific

    // Process argument(s) from the kernel
    cri_get_siginfo(info, siginfo);

    // Simply enter the PENDING state
    // The thread will then wake up and run the callback(s)
    if (cri_cmp_swap(&info->cr_state, CR_STATE_IDLE, CR_STATE_PENDING)) {
	cri_atomic_inc(&cri_live_count);
    }

    errno = saved_errno;
}

// thread_reset()
//
// This is a pthread_atfork callback to reset the state in the child process
static void thread_reset(void)
{
	pthread_mutex_init(&thread_lock, NULL);
	pthread_cond_init(&thread_init_cond, NULL);
	thread_state = CRI_THREAD_STOPPED;
}

// thread_main()
//
// This is the main loop of a thread running thread-context callbacks
static void* thread_main(void* arg)
{
#if LIBCR_TRACING
    int pid = (int)getpid();
#endif
    cri_info_t *info;
    sigset_t mask;
    int rc;

    // Reduce false wake-ups by blocking all signals.
    sigfillset(&mask);
    pthread_sigmask(SIG_SETMASK, &mask, NULL);

    // Tell the kernel code we are a Phase I thread.
    //
    // If we race against a checkpoint request, then after the
    // checkpoint has been taken, the cr_syscall() fails w/
    // (errno == EAGAIN) and we try again.
    do {
	rc = cri_syscall(CR_OP_HAND_PHASE1, &my_handler);
    } while ((rc < 0) && (errno == EAGAIN));
    if (rc != 0) {
	CRI_ABORT("CR_OP_HAND_PHASE1 failed w/ errno = %d", errno);
    }

    // Initialize our cri_info as a PHASE1 thread.
    // The write to 'handler' is naturally an atomic operation.
    //
    // If a request arrives any time before 'info->handler'
    // is set, then the kernel and/or the signal handler will still
    // think we are a PHASE II thread and we'll respond to the request
    // from signal context.  This is acceptable since we cannot possibly
    // have any callbacks registered until after we signal the condition
    // variable and subsequently release the thread_lock.
    info = cri_info_init();
    info->register_cmd = CR_OP_HAND_PHASE1;
    // XXX: Must write register_cmd before handler.  Need memory barrier?
    info->handler = &my_handler;

#if 1
    // XXX: as long as restart-time re-registration is done in user-space
    // there is a potential race condition if we are checkpointed between
    // the time we register with the kernel and the time we set the
    // info->handler.  In that case we won't get re-registered at
    // restart.  Here is a redundant registration to work around that:
    do {
	rc = cri_syscall(CR_OP_HAND_PHASE1, &my_handler);
    } while ((rc < 0) && (errno == EAGAIN));
    // Expect EEXIST in the normal case
    if ((rc != 0) && (errno != EEXIST)) {
	CRI_ABORT("redundant CR_OP_HAND_PHASE1 failed w/ errno = %d", errno);
    }
#endif

    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] started checkpoint thread",
		(int)getpid());

    // Wakeup the thread(s) blocked waiting for our initialization.
    // They can register callbacks only after we release thread_lock.
    pthread_mutex_lock(&thread_lock);
    thread_info_p = info;
    thread_state = CRI_THREAD_RUNNING;
    pthread_cond_broadcast(&thread_init_cond);
    pthread_mutex_unlock(&thread_lock);

    while (1) {
	// Block until the next request (or pthread cancellation) arrives
	(void)cri_syscall(CR_OP_HAND_SUSP, NULL);
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] checkpoint thread woke up", pid);

	// Check for cancellation.
	pthread_testcancel();

	// Get locks
	cri_black_lock(&cri_cs_lock);
	pthread_mutex_lock(&thread_lock);

	// Respond to a request iff one is pending, does nothing otherwise.
	cri_start_checkpoint(info, 1);

	// Release locks
	pthread_mutex_unlock(&thread_lock);
	cri_black_unlock(&cri_cs_lock);
    }

    // NOT REACHED
    return NULL;
}

// thread_init()
// 
// Ensure that the thread is started exactly once.
// Note that more than one thread might get here between the time the
// first enters pthread_cond_wait(), thus dropping the lock, and the
// time the thread gets far enough through its initialization
// to acquire the lock and signal the condition variable.
// 
// Called w/ thread_lock held.
static void thread_init(void)
{
    if (thread_state != CRI_THREAD_RUNNING) {
	if (thread_state == CRI_THREAD_STOPPED) {
	    int rc;

	    if (!cri_pthread_present) {
		CRI_ABORT("cr_register_callback() called w/ CR_THREAD_CONTEXT "
			  "but w/o libpthread linked");
	    }

	    rc = pthread_atfork(NULL, NULL, &thread_reset);
	    if (rc != 0) {
		CRI_ABORT("pthread_atfork() returned %d", rc);
	    }

	    rc = (*cri_real_pthread_create)(&my_thread, NULL,
					    &thread_main, NULL);
	    if (rc != 0) {
		CRI_ABORT("pthread_create() returned %d", rc);
	    }

	    thread_state = CRI_THREAD_STARTING;
	}

	// Wait for the thread to complete initialization before returning.
	while (thread_state != CRI_THREAD_RUNNING) {
	    pthread_cond_wait(&thread_init_cond, &thread_lock);
	}
    }
}

//
// Internal functions
//

// Not reentrant, but thread safe.
cr_callback_id_t
cri_register_thread(cri_info_t		*info,
		    cr_callback_t	func,
		    void*		arg,
		    int			flags)
{
    cr_callback_id_t	result;

    pthread_mutex_lock(&thread_lock);
    thread_init();
    result = cri_do_register(thread_info_p, func, arg, flags);
    pthread_mutex_unlock(&thread_lock);

    return result;
}

// Not reentrant, but thread safe.
int
cri_replace_thread(cri_info_t		*info,
		   cr_callback_id_t	id,
		   cr_callback_t	func,
		   void*		arg,
		   int			flags)
{
    int retval;

    pthread_mutex_lock(&thread_lock);
    retval = cri_do_replace(thread_info_p, id, func, arg, flags);
    pthread_mutex_unlock(&thread_lock);

    return retval;
}
