/* 
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_core.c,v 1.112 2003/11/14 19:55:24 phargrov Exp $
 */

#include <sys/ioctl.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <features.h>

// XXX: needed as long as we do some save/restore work in user space:
#include <termios.h>	// XXX: for termios
#include <sys/param.h>	// XXX: for getcwd() call
#include <sys/time.h>	// XXX: for set/getitimer

#include "cr_private.h"

//
// Private variables
//

//
// Public variables
//

// Holds the cri_info for the main (or only) thread
cri_info_t *cri_info_main_p;

// How many threads are actively trying to checkpoint (PENDING + ACTIVE)
cri_atomic_t cri_live_count = {0,};

// Red-black lock for critical sections
cri_rb_lock_t cri_cs_lock = CRI_RB_LOCK_INITIALIZER;

//
// Private functions
//

// Tell the kernel we are ready to be saved
// This version just invokes CR_OP_HAND_CHKPT
static int do_checkpoint(int flags)
{
    static cr_spinlock_t cwd_lock = CR_SPINLOCK_INITIALIZER;
    static char cwd[PATH_MAX];

    int rc;
#if CRI_DO_TERMIOS
    struct termios ti;
#endif
    char *cwdvalid;
    struct itimerval    real_timer;
    struct itimerval    virt_timer;
    struct itimerval    prof_timer;

    // "GLOBAL" CHECKPOINT STUFF HERE:

    // Save the itimers
    getitimer(ITIMER_REAL,    &real_timer);
    getitimer(ITIMER_VIRTUAL, &virt_timer);
    getitimer(ITIMER_PROF,    &prof_timer);

    // Save the cwd to a shared space.
    //
    // Note that getcwd() is not actually on the list of
    // signal-safe functions, so this should be in the kernel.
    //
    // We can't use an automatic variable because that would overflow the
    // tiny stack of the pthread manager.  Because there is a single
    // shared array we need to ensure that at most one thread is doing
    // this at once, but don't care how many eventually do so as long as
    // at least one does.  This construct fits the bill:
    if (cr_spinlock_trylock(&cwd_lock)) {
	cwdvalid = getcwd(cwd, PATH_MAX);
    } else {
	cwdvalid = NULL;
    }

#if CRI_DO_TERMIOS
    // Save the termios structure
    do {
	int tty = __libc_open("/dev/tty", O_WRONLY | O_NOCTTY);
	
	if (tty >= 0) {
	    rc = tcgetattr(tty, &ti);
	    if (rc < 0) {
		CRI_ABORT("Trouble saving termios");
	    }
	    __libc_close(tty);
	}
    } while (0);
#endif

#if CRI_PTHREAD_HACK
    cri_pthread_hack_checkpoint();
#endif /* CRI_PTHREAD_HACK */

    // XXX: This is needed until the kernel reopens the c/r fd:
    cri_fd_checkpoint();

    // This does the actual kernel call to take the checkpoint:
    rc = cri_syscall(CR_OP_HAND_CHKPT, (void *)flags);
    if (rc < 0) {
	return -errno;
    }

    if (rc) {
	// "GLOBAL" RESTART STUFF HERE

#if CRI_PTHREAD_HACK
	// XXX: Rebuild pthread pid tables and control pipe
	cri_pthread_hack_restart();
#endif

	// XXX: This is needed until the kernel reopens the c/r fd:
	cri_fd_restart();

#if CRI_DO_TERMIOS
	// Restore the termios structure
	do {
	    int tty = __libc_open("/dev/tty", O_WRONLY | O_NOCTTY);
	
	    if (tty >= 0) {
		int err = tcsetattr(tty, TCSANOW, &ti);
		if (err < 0) {
		    CRI_ABORT("Trouble restoring termios");
		}
		__libc_close(tty);
	    }
	} while (0);
#endif

	// Restore the cwd
	if (cwdvalid) {
	   int err = chdir(cwd);
	   if (err < 0) {
		CRI_ABORT("Trouble restoring working directory");
	   }
	}

	// Restore the itimers
	setitimer(ITIMER_REAL,    NULL, &real_timer);
	setitimer(ITIMER_VIRTUAL, NULL, &virt_timer);
	setitimer(ITIMER_PROF,    NULL, &prof_timer);
    } else {
	// "GLOBAL" CONTINUE STUFF HERE

#if CRI_PTHREAD_HACK
	cri_pthread_hack_continue();
#endif

	// XXX: This is needed until the kernel reopens the c/r fd:
	cri_fd_continue();
    }

    if (cwdvalid) {
	cr_spinlock_unlock(&cwd_lock);
    }

    return rc;
}

static int abort_checkpoint(int flags)
{
    int rc;

    rc = cri_syscall(CR_OP_HAND_ABORT, (void *)flags);

    return rc;
}

// enter_idle_state(info)
//
// Transition to CR_STATE_IDLE, regardless of starting state
static inline void enter_idle_state(cri_info_t *info)
{
    cri_atomic_write(&info->cr_state, CR_STATE_IDLE);
    cri_atomic_write(&info->cr_cs_count, 1);
}

// Signal handler to dispatch the request
//
// XXX: This will eventually all be done by the kernel which will
// invoke the proper handler directly w/o using a signal number.
static void cri_sig_handler(int signr, siginfo_t *siginfo, void *context)
{
    int saved_errno = errno;
#if LIBCR_TRACING
    int pid = (int)getpid();
#endif
    cri_info_t *info = cri_info_location();	// thread-specific

    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] signal received", pid);
    LIBCR_TRACE(LIBCR_TRACE_INFO,
		"checkpoint will deliver signal %d\n", siginfo->si_pid);

    if (!info || !info->handler) {
	int rc;

	// No state and/or handler, so do this the "easy way"
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] START", pid);
	rc = do_checkpoint(0);
	if (rc < 0) {
	    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] do_checkpoint failed", pid);
	}
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] DONE", pid);
    } else {
	// Run the handler appropriate to this thread
	info->handler(signr, siginfo, context);
    }

    errno = saved_errno;
}

//
// Public functions
//

// Request an application-initiated checkpoint
// The actual processing might be delayed by critical sections.
//
// Thread-safe, but not reentrant
void cr_request(void)
{
#define NAMELEN 16 /* context.XXXXX + \0 + extra */
    struct cr_chkpt_req req;
    char filename[NAMELEN];
    int fd;
    int rc;

    req.cr_target	= 0;		// 0 = self
    req.cr_scope	= CR_SCOPE_PROC;
    req.cr_secs		= 0;		// 0 = unbounded
    req.cr_opts		= NULL;
    req.cr_fd		= CR_DEST_CWD;
    req.dump_format	= cr_format_vmadump;
    req.signal		= 0;		// 0 = no signal
    req.flags		= 0;

#if 1 // Until VMAdump code deals w/ directory targets properly
    {
	snprintf(filename, NAMELEN, "context.%d", (int)getpid());
	fd = __libc_open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0644);
	if (fd < 0) {
	    CRI_ABORT("open(%s)", filename);
	}
	req.cr_fd = fd;
    }
#endif

    // If there is a previous request un-reaped then reap it now.
    do {
	rc = cri_syscall(CR_OP_CHKPT_REAP, CRI_SYSCALL_NOARG);
	if (rc < 0) {
	    if (errno == EINVAL) {
		// No request was pending - not an error.
		break;
	    } else if (errno == EAGAIN) {
		// Wait for the request to complete
		// If interrupted by a signal the loop will try again
		cri_syscall(CR_OP_CHKPT_DONE, NULL);
	    } else {
		// XXX: other error cases?
	    }
	}
    } while (rc < 0);

    rc = cri_syscall(CR_OP_CHKPT_REQ, &req);
    if (rc) {
	CRI_ABORT("CR_OP_CHKPT_REQ returned %d w/ errno=%d\n", rc, errno);
    }

    // Don't return until the request is either pending or completed.
    // This is important; otherwise we would have races with critical
    // sections which follow the cr_request() call.
    //
    // We can detect pending state because the live count is non-zero.
    // We can detect completed because a reap will succeed (or fail
    // with an errno other than EAGAIN).
    // When restarting we expect the reap to fail w/ errno = EINVAL,
    // (if the fd is restored, since there won't be any checkpoint
    // request anymore) or with EBADF if the fd is not restored.
    rc = 0;
    CRI_WHILE_COND_YIELD(
	!cri_atomic_read(&cri_live_count) &&
	(rc = cri_syscall(CR_OP_CHKPT_REAP, CRI_SYSCALL_NOARG)) != 0 &&
	errno == EAGAIN
    );

    /* Silently ignore any failure here */
    (void)close(fd);

    /* Remove checkpoint file if an unexpected error occurred (see above) */
    if (rc && (errno != EAGAIN) && (errno != EINVAL) && (errno != EBADF)) {
	LIBCR_TRACE(LIBCR_TRACE_INFO, "Removing supposedly bogus context file "
		    "(%s), because rc=%d and errno=%d (%s)",
		    filename, rc, errno, strerror(errno));
	if ( (rc = unlink(filename))) {
	    CRI_ABORT("unable to remove aborted checkpoint file %s: %s\n", 
		      filename, cr_strerror(errno));
	}
    }
}

int cr_checkpoint(int flags)
{
    cri_info_t *info = cri_info_location();	// thread-specific
    int index;
    int retval;
    int rc;

    // Check for improper calls
    if (cri_atomic_read(&info->cr_state) != CR_STATE_ACTIVE) {
	CRI_ABORT("STATE ERROR.  Called while not in the ACTIVE state.");
    }

    // Track call depth
    info->run.depth += 1;

#if LIBCR_TRACING
	int pid = (int)getpid();
#endif
    // Loop calling callbacks until they have all been called and then
    // invoke do_checkpoint() to do the real work.
    //
    // This strange loop ensures that after the callback with (index == 0)
    // we next call do_checkpoint() even if callback0 is NULL or returns
    // without calling back into cr_checkpoint().
    do {
	if (flags & (CR_CHECKPOINT_TEMP_FAILURE|CR_CHECKPOINT_PERM_FAILURE)) {
	    // We've detected failure of the invoking callback
	    // We we must now call the kernel and stop stacking of callbacks.
	    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] CANCELLING CHECKPOINT (flags=%d)",
			pid, flags);
	    if (abort_checkpoint(flags) == -1)
		info->run.rc = -errno;
	    index = info->run.index = -1;
	    break;
	}
    
	index = --info->run.index;
	if (index >= 0) {
	    // Call the next callback, skipping NULLs

	    if (info->cr_cb[index].func != NULL) {
		rc = (*info->cr_cb[index].func)(info->cr_cb[index].arg);
		if (rc) {
		    CRI_ABORT("Callback %d returned %d - ABORTING\n",
			      index, rc);
		}
	    }
	} else {
	    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] CHECKPOINTING (flags=%d)",
			pid, flags);
	    info->run.rc = do_checkpoint(flags);
	    if (info->run.rc < 0) {
		LIBCR_TRACE(LIBCR_TRACE_INFO,
			    "[%d] do_checkpoint() FAILED", pid);
	    } else if (info->run.rc) {
		LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] RESTORING", pid);
	    } else {
		LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] CONTINUING", pid);
	    }
	}
    } while (info->run.index >= 0);

    // Once the outer-most invocation calls CR_OP_HAND_{CONT,RSTRT}
    // the kernel might checkpoint us again.  Save the return value now
    // because a second checkpoint could modify info->run.rc.
    // We take this measure even though the final return is currently
    // ignored by cri_start_checkpoint().
    retval = info->run.rc;

    // Perform cleanup only on the final/outer-most return
    if (--info->run.depth == 0) {
	enter_idle_state(info);
	cri_atomic_dec_and_test(&cri_live_count);
	if (retval > 0) {
	    rc = cri_syscall(CR_OP_HAND_RSTRT, CRI_SYSCALL_NOARG);
#if 0	//XXX: unimplemented, so expect an error
	    if (rc < 0) {
		CRI_ABORT("cri_syscall(CR_OP_HAND_RSTRT) failed");
	    }
#endif
	} else {
	    rc = cri_syscall(CR_OP_HAND_CONT, CRI_SYSCALL_NOARG);
	    if (rc < 0) {
		CRI_ABORT("cri_syscall(CR_OP_HAND_CONT) failed");
	    }
	}
    }

    return retval;
}

//
// Internal functions
//

// cri_checkpoint_info_init()
//
// Initialize a thread-local struct cr_checkpoint_info
void cri_get_siginfo(cri_info_t *info, siginfo_t *siginfo)
{
    info->cr_checkpoint_info.signal = siginfo->si_pid;
}


// cri_info_init()
//
// Routine to process the list of registered callbacks
// On first call per thread, allocates the thread-specific cri_info structure.
cri_info_t* cri_info_init(void)
{
    cri_info_t *info = cri_info_location();	// thread-specific
    int rc;

    if (!info) {
	info = malloc(sizeof(cri_info_t));
	if (!info) {
	    CRI_ABORT("Failed to allocate cri_info structure");
	}

	memset(info, 0, sizeof(cri_info_t));
	enter_idle_state(info);
	info->handler = NULL;
	info->next_id = 0;

	if (cri_pthread_present) {
	    rc = pthread_setspecific(cri_info_key, info);
	    if (rc != 0) {
		CRI_ABORT("pthread_setspecific() returned %d", rc);
	    }
	} else {
	    cri_info_main_p = info;
	}
    }

    return info;
}

// cri_start_checkpoint()
//
// Routine to process the list of registered callbacks
void cri_start_checkpoint(cri_info_t *info, int is_thread)
{
    if (cri_cmp_swap(&info->cr_state, CR_STATE_PENDING, CR_STATE_ACTIVE)) {
	// State is valid, so start running callbacks
#if LIBCR_TRACING
	int pid = (int)getpid();
#endif
	info->run.depth = 0;
	info->run.index = info->cr_cb_count;
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] START", pid);
	(void)cr_checkpoint(0);
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] DONE", pid);
    } else if (!is_thread) {
	CRI_ABORT("STATE ERROR.  Probably an excess call to cr_leave_cs().");
    } else {
	// False wake up of a checkpoint thread - not an error.
    }
}

/* Initialization logic.
 *
 * XXX: need a balancing _fini()?
 */
static int raninit = 0;

static void _do_init(void)
{
    int rc;
    struct sigaction	sa;
    int signum;

    //
    // Initialize tracing
    //
    libcr_trace_init();

    //
    // Initialize pthread-dependent parts if appropriate
    //
    cri_pthread_init();

    //
    // Setup signal handler
    //
    signum = __libc_allocate_rtsig(0);
    if (signum != CR_SIGNUM) {
	CRI_ABORT("Failed to allocate signal %d in process %d: got signal %d instead",
		  CR_SIGNUM, (int)getpid(), signum);
    }
    sa.sa_sigaction = cri_sig_handler;
    sa.sa_flags = SA_RESTART | SA_SIGINFO;
    rc = sigfillset(&sa.sa_mask);
    if (rc != 0) {
	CRI_ABORT("sigfillset() failed: %s", strerror(errno));
    }
#if HAVE___LIBC_SIGACTION
    rc = __libc_sigaction(CR_SIGNUM, &sa, NULL);
#elif HAVE___SIGACTION
    rc = __sigaction(CR_SIGNUM, &sa, NULL);
#else
#   error "Don't know how to invoke sigaction() bypassing libpthread"
#endif
    if (rc != 0) {
	CRI_ABORT("sigaction() failed: %s", strerror(errno));
    }
}

/* Initialization entry point.
 *
 * If shared lib, compiled with -DPIC, and use magic '_init' function, which
 * is always loaded just once.
 *
 * If compiled statically, use the 'constructor' attribute to run at
 * startup.
 *
 * TODO:  This doesn't completely solve the problem:  
 *  1)  When statically linked, if no other functions from this library used,
 *	linker won't link the library into the app, and so constructor not
 *	run.  
 *  2) If statically linked AND started with cr_run, init function will be run
 *  twice (and with dynamic lib having its own copy of all data structures).
 *	- calls from executable to libcr will be serviced by static copy of
 *	  lib, while those from libraries that linked against libcr.so will be
 *	  serviced by dynamic lib.  
 *	    - might be able to work things out if all data structures stored
 *	      in thread-local data--either lib's signal handler could find all
 *	      data
 *	    - Or, could put burden on users not to link both.
 *	    - But for now we're 'solving' issue by only providing libcr.so
 */
#ifdef PIC
void _init(void)
{
    if (!raninit++)
	_do_init();
}
#else
# error PIC not defined  (note: static libcr.a not yet supported)
#if 0
static void __attribute__((constructor)) _cr_static_init(void)
{
    if (!raninit++)
	_do_init();
}
#endif /* 0 */
#endif

