/* 
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_syscall.c,v 1.6 2003/11/14 19:55:24 phargrov Exp $
 *
 *
 * This file abstracts away the ioctl() interface so that we
 * can painlessly switch to a "real" syscall someday.
 */

#include <sys/ioctl.h>	// for ioctl()
#include <fcntl.h>	// for O_WRONLY
#include <stdio.h>	// for fprintf, stderr
#include <errno.h>	// for errno
#include <unistd.h>	// for dup2

#include "cr_private.h"

//
// Private data
//

static cri_atomic_t	local_fd	= {(unsigned long)(-1L)};
static cr_spinlock_t	local_lock	= CR_SPINLOCK_INITIALIZER;
static cri_atomic_t	local_barrier	= {0,};

//
// Private functions
//

static void
reregister_handler(void)
{
    cri_info_t *info = cri_info_location();	// thread-specific

    if (info && info->handler) {
	int fd, rc;

	// Know this to be valid when we get here
	fd = (int)cri_atomic_read(&local_fd);

	do {
	    rc = ioctl(fd, info->register_cmd, info->handler);
	} while ((rc < 0) && (errno == EAGAIN));

	if (rc < 0) {
	    if (errno == EEXIST) {
		fprintf(stderr, "Handler still registered at restart.  "
				"You should remove the work-around code "
				"in libcr/cr_syscall.c\n");
	    } else {
		CRI_ABORT("CR_OP_HAND_PHASE[12] failed w/ errno = %d", errno);
	    }
	}
    } else {
	// NOTHING TO DO
    }
}


//
// Public Functions
//

// Open a single, shared, instance of the control node.
// Note that this is signal and thread safe.
//
// returns < 0 on failure
int cri_get_fd(void)
{
    int fd;

    while ((fd = (int)cri_atomic_read(&local_fd)) < 0) {
	int old = fd;

	fd = __libc_open(CR_CTRL_FILE, O_WRONLY);
	if (fd < 0) {
	    // Fail
	    break;
	} else if (!cri_cmp_swap(&local_fd, old, fd)) {
	    // We lost the race: another thread has opened the file
	    __libc_close(fd);  // may fail silently
	} else {
	    // Success
	    break;
	}
    }

    return fd;
}

int cri_syscall(unsigned long op, void * arg)
{
    int fd = cri_get_fd();

    if (fd < 0) {
	CRI_ABORT("open(" CR_CTRL_FILE ")");
	/* Not reached */
    } 

    return ioctl(fd, op, arg);
}

void cri_fd_checkpoint(void)
{
    cri_atomic_inc(&local_barrier);
}

void cri_fd_continue(void)
{
    cri_atomic_write(&local_barrier, 0);
}

void cri_fd_restart(void)
{
    int fd, rc;

    if (cr_spinlock_trylock(&local_lock)) {
	// EXACTLY ONE THREAD RUNS HERE
	int saved;

	// Get the saved value
	saved = (int)cri_atomic_read(&local_fd);
	if (saved < 0) {
	    CRI_ABORT("No saved fd to restore");
	    return;	/* NOT REACHED */
	}

	// Reopen the ctrl node
	fd = __libc_open(CR_CTRL_FILE, O_WRONLY);
	if (fd < 0) {
	    CRI_ABORT("open(" CR_CTRL_FILE ")");
	    return;	/* NOT REACHED */
	}

	// Use dup2() to get the desired descriptor
	if (fd != saved) {
	    rc = dup2(fd, saved);
	    if (rc < 0) {
		CRI_ABORT("dup2(" CR_CTRL_FILE ")");
		return;	/* NOT REACHED */
	    }
	    __libc_close(fd);  // may fail silently
	}

	cri_barrier_enter(&local_barrier);
	cr_spinlock_unlock(&local_lock);
    } else {
	// Make sure all other threads wait until fd is restored
	cri_barrier_enter(&local_barrier);
    }

    // XXX: work-around to reregister w/ kernel after restart:
    reregister_handler();
}
