/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_restart.c,v 1.15 2003/11/14 19:55:41 phargrov Exp $
 */


#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <getopt.h>

#include <sys/wait.h>
#include <sched.h>
#include <errno.h>

#include "cr.h"

/* can't use argv[0] to get name, since libtool screws it up */
#define MY_NAME	"cr_restart"

static void usage(FILE *stream)
{
    fprintf(stream,
"Usage: " MY_NAME " checkpoint_file\n"
"Options:\n"
" -?, --help	    print this help message.\n"
"     --version	    print version information.\n" 
    );
}

sig_atomic_t child = 0; /* who did we spawn? */
static void signal_child (int, siginfo_t *, void *);

static void signal_self(int sig)
{
    struct sigaction sa;

    /* restore default (in kernel) handler */
    sa.sa_handler = SIG_DFL;
    sa.sa_flags = SA_RESTART | SA_NOMASK;
    sigemptyset(&sa.sa_mask);
    (void)sigaction(sig, &sa, NULL);

    /* send to self */
    raise(sig);

    /* restore self as handler */
    sa.sa_sigaction = &signal_child;
    sa.sa_flags = SA_RESTART | SA_NOMASK | SA_SIGINFO;
    (void)sigaction(sig, &sa, NULL);
}

/* Best-effort to forward all signals to the restarted child
 * XXX: This is probably not complete in terms of special-cases
 */
static void signal_child (int sig, siginfo_t *siginfo, void *context)
{
    if ((siginfo->si_code == SI_KERNEL) &&
	(sig == SIGILL || sig == SIGFPE || sig == SIGBUS || sig == SIGSEGV )) {
	/* This signal is OUR error, so we don't forward */
	signal_self(sig);
    } else if (sig == SIGTSTP || sig == SIGTTIN || sig == SIGTTOU) {
	/* The catchable stop signals go to child AND self */
	(void)kill(child, sig);
	signal_self(sig);
    } else {
	/* default case */
	(void)kill(child, sig);
    }
}

static void print_version()
{
    printf(MY_NAME " version %s\n", CR_RELEASE_VERSION);
}

/* getopt enum */
enum {
    opt_version
};

/* try to exit such that our exit code is same as our child */
static void mimic_exit(int status)
{
    if (WIFEXITED(status)) {
	/* easy to mimic normal return */
	exit(WEXITSTATUS(status));
    } else if (WIFSIGNALED(status)) {
	/* disable generation of a 'core' */
	struct rlimit r;
	r.rlim_cur = r.rlim_max = 0;
	(void)setrlimit(RLIMIT_CORE, &r);

	/* now raise the signal */
	signal_self(WTERMSIG(status));
    } else {
	fprintf(stderr, "Unexpected status from child\n");
	exit(-1);
    }
}

int main(int argc, char **argv)
{
    struct cr_rstrt_args req;
    int context_fd;
    int proc_fd;
    int err;
    char *context_filename;
    char error_message[255];
    struct sigaction sa;
    sigset_t sigmask, oldmask;
    int sig;
    int child_status;

    char * shortflags = "h?";  /* 1 colon == requires argument */
    struct option longflags[] = {
	/* misc options: */
	{ "help",    no_argument,       0, '?' },
	{ "version", no_argument,       0, opt_version },
	{ 0,	     0,		        0,  0  }
    };

    while (1) {
	int longindex = -1;
	int opt = getopt_long(argc, argv, shortflags, longflags, &longindex);

	if (opt == -1)
	    break;		    /* reached last option */

	switch (opt) {
	    case opt_version:
		print_version();
		exit(0);
	    case '?':
	    case 'h':
		usage(stdout);
		exit(0);
	    default:
		usage(stderr);
		exit(-1);
	}
    }

    if (argc != 2) {
        usage(stderr);
        return -1;
    } else {
      context_filename = argv[1];
    }

    /* ... open up the control file */
    proc_fd = open(CR_CTRL_FILE, O_WRONLY);
    if (proc_fd < 0) {
        perror("open(" CR_CTRL_FILE ")");
        return proc_fd;
    }

    /* ... open up the context file */
    context_fd = open(context_filename, O_RDONLY);
    if (context_fd < 0) {
        /* XXX:  Clean this up to be consistent */
        snprintf(error_message, sizeof(error_message), 
          "open(%s, O_RDONLY)", context_filename);
        perror(error_message);
        return context_fd;
    }

    /* ... initialize the request structure */
    req.cr_fd       = context_fd;
    req.cr_opts     = NULL;

    /* ... tell the kernel to go off and restart something */
    err = ioctl(proc_fd, CR_OP_RSTRT_REQ, &req);
    if (err < 0) {
        /* The restart request was unsuccessful.  */
        perror("ioctl(" CR_CTRL_FILE ", CR_OP_RSTRT_REQ)");
        return err;
    } else if (err == 0) {
        /* If we're in here, we're a dying process, soon to be reborn
         * as the restarting job. This ioctl will never return. */
        err = ioctl(proc_fd, CR_OP_RSTRT_CHILD, NULL);
        if (err == 0) {
            err = ioctl(proc_fd, CR_OP_RSTRT_CHILD, NULL);
            perror("unexpeted return from ioctl(" CR_CTRL_FILE ", CR_OP_RSTRT_CHILD #2)");
	} else {
            perror("ioctl(" CR_CTRL_FILE ", CR_OP_RSTRT_CHILD #1)");
	}
	return err;
    }

    /* ... wait forever for the restart to complete */
    err = ioctl(proc_fd, CR_OP_RSTRT_DONE, NULL);
    if (err < 0) {
        /* There was a problem with the restart.  */
        perror("ioctl(" CR_CTRL_FILE ", CR_OP_RSTRT_DONE)");
	return err;
    }

    /* ... prepare to forward signals */
    /* XXX: what should we use in place of _NSIG? */
    sigfillset(&sigmask);
    sigdelset(&sigmask, SIGSTOP);
    sigdelset(&sigmask, SIGKILL);
    sigdelset(&sigmask, SIGCHLD);
    sigprocmask(SIG_SETMASK, &sigmask, &oldmask);
    sa.sa_sigaction = &signal_child;
    sa.sa_flags = SA_RESTART | SA_NOMASK | SA_SIGINFO;
    sigemptyset(&sa.sa_mask);
    for (sig = 1; sig < _NSIG; ++sig) {
	if (sig == SIGCHLD) continue;
	if (sig == SIGKILL) continue;
	if (sig == SIGSTOP) continue;

	/* best-effort so we ignore failures */
	(void)sigaction(sig, &sa, NULL);
    }

    /* ... reap the checkpoint, collecting a child pid on success */
    err = ioctl(proc_fd, CR_OP_RSTRT_REAP);
    if (err < 0) {
        /* Unable to reap! */
        perror("ioctl(" CR_CTRL_FILE ", CR_OP_RSTRT_REAP)");
	return err;
    } else {
	child = (pid_t)err;
    }

    /* ... now we forward the signals */
    sigprocmask(SIG_SETMASK, &oldmask, NULL);

    err = close(proc_fd);
    if (err < 0) {
        perror("close(" CR_CTRL_FILE ")");
        return err;
    }

    /* XXX should have the option of a non-blocking case */
    do {
        /* wait for the child we've inherited */
        err = waitpid(child, &child_status, __WCLONE | __WALL);
    } while ((err < 0) && (errno == EINTR)); /* retry on non-fatal signals */
    if (err < 0) {
        /* Unable to reap the child */
        perror("waitpid()");
	return err;
    }	

    mimic_exit(child_status);
    return -1; /* NOT REACHED */
}
